using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using Equis.Xna.Input;
using System.Diagnostics;
using System.Threading;

namespace Demos.Calibrate {
    public class Particle {
        public Vector2 Position;
        public Vector2 Velocity;
        public Color StartColor;
        public Color EndColor;
        public float StartRadius;
        public float EndRadius;
        public float StartTime;
        public float LifeTime;
    };

    public class CalibrateGame : Microsoft.Xna.Framework.Game {
        private readonly Random random = new Random();

        private readonly GraphicsDeviceManager manager;
        private readonly Ftir ftir;

        private SpriteBatch batch;
        private Texture2D background;
        private Texture2D circle;

        private bool showOutput;
        private bool drift;

        private KeyboardState oldKeyboard;

        private List<Particle> particles = new List<Particle>();

        private List<Vector2> desired = new List<Vector2>();
        private List<Vector2> reported = new List<Vector2>();
        private int target = -1;

        public CalibrateGame() {
            manager = new GraphicsDeviceManager(this);
            manager.IsFullScreen = true;

            ftir = Ftir.FromCapture(null);

            Content.RootDirectory = "Content";
        }

        protected override void Initialize() {
            base.Initialize();
        }

        protected override void LoadContent() {
            batch = new SpriteBatch(GraphicsDevice);
            background = new Texture2D(GraphicsDevice, 1024, 512,
                1, TextureUsage.AutoGenerateMipMap, SurfaceFormat.Color);

            circle = Content.Load<Texture2D>("Textures/Circle");
        }

        protected override void UnloadContent() {
        }

        protected override void Update(GameTime now) {

            float elapsed = (float)now.TotalGameTime.TotalSeconds;

            KeyboardState newKeyboard = Keyboard.GetState();

            if (newKeyboard.IsKeyDown(Keys.Escape)) {
                Environment.Exit(0);
            }

            if (oldKeyboard.IsKeyUp(Keys.C) && newKeyboard.IsKeyDown(Keys.C)) {
                desired.Clear();
                reported.Clear();
                int width = GraphicsDevice.Viewport.Width;
                int height = GraphicsDevice.Viewport.Height;

                int size = 3;
                int xpad = 128;
                int ypad = 128;
                for (int row = 0; row <= size; ++row) {
                    for (int col = 0; col <= size; ++col) {
                        int x = xpad + (width - 2 * xpad) * row / size;
                        int y = ypad + (height - 2 * ypad) * col / size;
                        desired.Add(new Vector2((float)x / width, (float)y / height));
                    }
                }

                target = 0;
                ftir.Calibrate(null, null);
            }

            if (oldKeyboard.IsKeyUp(Keys.D) && newKeyboard.IsKeyDown(Keys.D)) {
                drift = !drift;
            }
            if (oldKeyboard.IsKeyUp(Keys.I) && newKeyboard.IsKeyDown(Keys.I)) {
                showOutput = false;
            }
            if (oldKeyboard.IsKeyUp(Keys.O) && newKeyboard.IsKeyDown(Keys.O)) {
                showOutput = true;
            }

            if (oldKeyboard.IsKeyUp(Keys.F) && newKeyboard.IsKeyDown(Keys.F)) {
                manager.ToggleFullScreen();
            }

            oldKeyboard = newKeyboard;

            if (showOutput) {
                ftir.PreviewFiltered(background);
            } else {
                ftir.PreviewInput(background);
            }

            FtirEvent[] events = ftir.PollEvents(0.0f);

            FtirEvent calibrate = null;
            bool ambiguous = false;

            foreach (FtirEvent e in events) {
                if (e.Since + 15 > e.Frame) {
                    if (calibrate == null)
                        calibrate = e;
                    else
                        ambiguous = true;
                }

                //if (random.NextDouble() * particles.Count > 50) continue;

                float scale = (float)Math.Sqrt(e.Mass);
                if (e.Press || e.Release) {
                    scale = Math.Max(scale, 1000.0f);
                }

                Particle p = new Particle();
                p.Position = e.Position;
                //new Vector2(0.0f, scale * -1e-4f)
                p.Velocity = drift ? new Vector2(0.0f, -0.2f) : Vector2.Zero;

                if (e.Press) {
                    p.StartColor = new Color(0.4f, 1.0f, 0.4f, 0.75f);
                } else if (e.Release) {
                    p.StartColor = new Color(0.0f, 0.6f, 1.0f, 0.75f);
                } else {
                    p.StartColor = new Color(1.0f, 0.8f, 0.0f, 0.5f);
                }
                p.EndColor = new Color(p.StartColor, 0.0f);
                p.StartRadius = scale / 35.0f;
                p.EndRadius = p.StartRadius / 4;
                p.StartTime = elapsed;
                p.LifeTime = 2.0f;

                particles.Add(p);
            }

            if (target >= 0 && calibrate != null && !ambiguous) {
                if (reported.Count > 0 && Vector2.Distance(reported[target - 1], calibrate.Position) < 0.05) {
                    // This point is too close to the previous one.
                } else if (target != reported.Count) {
                    throw new InvalidOperationException("Calibration out of order.");
                } else {
                    reported.Add(calibrate.Position);
                    target += 1;
                    if (target >= desired.Count) {
                        ftir.Calibrate(desired.ToArray(), reported.ToArray());
                        target = -1;
                    }
                }
            }

            base.Update(now);
        }

        protected override void Draw(GameTime now) {
            float elapsed = (float)now.TotalGameTime.TotalSeconds;
            GraphicsDevice g = GraphicsDevice;

            g.Clear(Color.Black);

            Vector2 origin = new Vector2(circle.Width / 2.0f, circle.Height / 2.0f);

            batch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
            {
                Rectangle src = new Rectangle(0, 0, ftir.Format.Width, ftir.Format.Height);
                Rectangle dst = new Rectangle(0, 0, g.Viewport.Width, g.Viewport.Height);
                batch.Draw(background, dst, src, Color.White);
            }

            foreach (Particle p in particles) {
                float age = (elapsed - p.StartTime) / p.LifeTime;
                if (age > 1)
                    continue;

                Vector2 position = p.Position + p.Velocity * (elapsed - p.StartTime);
                position.X *= g.Viewport.Width;
                position.Y *= g.Viewport.Height;

                Color color = Color.Lerp(p.StartColor, p.EndColor, age);
                float scale = MathHelper.Lerp(p.StartRadius, p.EndRadius, age) / circle.Width;
                batch.Draw(circle, position, null, color, 0.0f, origin, scale, SpriteEffects.None, 0.0f);
            }

            if (target >= 0) {
                Vector2 position = desired[target];
                position.X *= g.Viewport.Width;
                position.Y *= g.Viewport.Height;

                batch.Draw(circle, position, null, Color.White, 0.0f, origin, 1.0f, SpriteEffects.None, 0.0f);
            }

            batch.End();

            base.Draw(now);
        }
    }
}
