#ifndef EQUIS_MACROS_HPP
#define EQUIS_MACROS_HPP

/// @name Utility macros
//@{

/// Indicates that a class does not expose a copy constructor.
///
/// By default, C++ will generate a copy constructor which performs a
/// per-field shallow copy. This macro may be placed in the top of a
/// class declaration to suppress that behavior.
///
/// This macro emits a private copy constructor. The constructor will
/// prevent C++ from generating any implicit constructors, so protects
/// against unexpected copies. The copy constructor declared by this
/// macro should not be implemented; if a private copy constructor is
/// desirable, declare it normally.

#define EQUIS_NO_COPY(Type) \
private: \
    Type(const Type& source);

/// Generates a unique name with the supplied prefix.
///
/// The generated name will always be a legal C++ identifier. Names may
/// collide across compilation units, so must be limited to file scope
/// -- for example, via 'static' or anonymous namespaces.

#define EQUIS_UNIQUE_NAME(Prefix) \
    Prefix ## __COUNTER__

/// Determines the static capacity of the specified array.
///
/// This macro is only valid for array parameters, not pointers.
#define EQUIS_ARRAY_CAPACITY(A) (sizeof(A)/sizeof(*(A)))

//@}

/// Enters a namespace.
#define EQUIS_ENTER_NS1(A) namespace A {

/// Leaves a namespace.
#define EQUIS_LEAVE_NS1(A) }

/// Enters a nested namespace.
#define EQUIS_ENTER_NS2(A,B) namespace A { namespace B {

/// Leaves a nested namespace.
#define EQUIS_LEAVE_NS2(A,B) } }

/// Enters a nested namespace.
#define EQUIS_ENTER_NS3(A,B,C) namespace A { namespace B { namespace C {

/// Leaves a nested namespace.
#define EQUIS_LEAVE_NS3(A,B,C) } } }

/// Includes types from an absolute namespace.
#define EQUIS_USING_NS1(A) using namespace ::A;

/// Includes types from an absolute namespace and its parents.
#define EQUIS_USING_NS2(A,B) EQUIS_USING_NS1(A); using namespace ::A::B;

/// Includes types from an absolute namespace and its parents.
#define EQUIS_USING_NS3(A,B,C) EQUIS_USING_NS2(A,B); using namespace ::A::B::C;

/// @name Debugging macros
///
/// - STATIC: Checked at compile-time.
/// - DEBUG: Checked when the code is executed in debug mode.
/// - (nothing): Checked when the code is executed.
///
/// Any failed runtime check will trigger the debugger.
/// 
/// - ASSERT: Terminate the program if the check fails.
/// - CHECK: Log a warning if the check fails.
///
//@{

#define EQUIS_STATIC_ASSERT(A) \
typedef int EQUIS_UNIQUE_NAME(_equis_static_assert)[(A) ? 1 : -1];

#define EQUIS_ASSERT(A) \
do { if (!(A)) { ::equis::assert_failed(__FILE__, __LINE__, #A); } } while (0)

#define EQUIS_CHECK(A) \
do { if (!(A)) { ::equis::check_failed(__FILE__, __LINE__, #A); } } while (0)

#if defined(DEBUG)
#define EQUIS_DEBUG_ASSERT(A) EQUIS_ASSERT(A)
#define EQUIS_DEBUG_CHECK(A) EQUIS_CHECK(A)
#else /* DEBUG */
#define EQUIS_DEBUG_ASSERT(A) do { } while (0)
#define EQUIS_DEBUG_CHECK(A)  do { } while (0)
#endif /* DEBUG */

#define EQUIS_ASSERT_NOT_NULL(A) EQUIS_ASSERT((A) != NULL)
#define EQUIS_CHECK_NOT_NULL(A) EQUIS_CHECK((A) != NULL)

//@}

#endif /* EQUIS_MACROS_HPP */
