#ifndef IN_EQUIS_MATH_HPP
#error This file must be included via "math.hpp"
#endif

EQUIS_ENTER_NS1(equis);

template < typename T > inline
T*
byte_offset(T* base, ptrdiff_t offset)
{
    char* p = reinterpret_cast<char*>(base) + offset;
    return reinterpret_cast<T*>(p);
}

template < typename T > inline
const T*
byte_offset(const T* base, ptrdiff_t offset)
{
    const char* p = reinterpret_cast<const char*>(base) + offset;
    return reinterpret_cast<const T*>(p);
}


template < typename T > inline 
int
count_one_bits(T value) {
    typedef typename numeric_traits<T>::assert_integer type_must_be_an_integer;
    typedef typename numeric_traits<T>::assert_unsigned type_must_be_unsigned;

    int a = 0;
    for (; value; value >>= 1) a += value & 0x1;
    return a;
}

template < typename T > inline
int
count_zero_bits(T value) {
    return count_one_bits(~value);
}

template < typename T > inline
int
leading_zero_bits(T value) {
    typedef typename numeric_traits<T>::assert_integer type_must_be_an_integer;
    typedef typename numeric_traits<T>::assert_unsigned type_must_be_unsigned;

    const int N = 8 * sizeof(T);
    for (int i = 0; i < N; ++i) {
        if (value >> (N-i)) return i;
    }
    return N;
}

template < typename T > inline
int
leading_one_bits(T value) {
    return leading_zero_bits(~value);
}

template < typename T > inline int
trailing_zero_bits(T value) {
    typedef typename numeric_traits<T>::assert_integer type_must_be_an_integer;
    typedef typename numeric_traits<T>::assert_unsigned type_must_be_unsigned;

    const int N = 8 * sizeof(T);
    for (int i = 0; i < N; ++i) {
        if ((value >> i) & 0x1) return i;
    }
    return N;
}

template < typename T > inline
int
trailing_one_bits(T value) {
    return trailing_zero_bits(~value);
}

template < typename T > inline
T
power_of_two(T value) {
    if (value == 0) return 0;
    return next_power_of_two(value - 1);
}

template < typename T > inline
T 
next_power_of_two(T value) {
    typedef typename numeric_traits<T>::assert_integer type_must_be_an_integer;
    typedef typename numeric_traits<T>::assert_unsigned type_must_be_unsigned;

    const int N = 8 * sizeof(T);
    for (int i = 1; i < N; i *= 2) {
        value |= value >> i;
    }
    return value + 1;
}

template < typename T > inline
T
mul_hi(T a, T b) {
    typedef typename numeric_traits<T>::assert_integer type_must_be_an_integer;
    typedef typename numeric_traits<T>::assert_unsigned type_must_be_unsigned;
    typedef typename numeric_traits<T>::promoted_type promoted_type;
    
    const int N = 8 * sizeof(T);
    promoted_type tp = static_cast<promoted_type>(a) * static_cast<promoted_type>(b);
    return static_cast<T>(tp >> N);
}

template < typename T > inline
typename numeric_traits<T>::promoted_type
mul(T a, T b) {
    typedef typename numeric_traits<T>::promoted_type promoted;
    return static_cast<promoted>(a) * static_cast<promoted>(b);
}

template < typename T > inline
T
align(T value, T mask) {
    if (!value || !mask) return value;
    value -= 1;
    value |= (mask - 1);
    value += 1;
    return value;
}

template < typename T > inline
T
clamp(T value, T minimum, T maximum) {
    if (value < minimum) return minimum;
    if (value < maximum) return value;
    return maximum;
}

template < typename T, typename Less > inline
T
clamp(T value, T minimum, T maximum, Less func) {
    if (func(value, minimum)) return minimum;
    if (func(value, maximum)) return value;
    return maximum;
}

EQUIS_LEAVE_NS1(equis);
