#include "common.hpp"
#include "View.hpp"
EQUIS_USING_NS1(equis);

__declspec(noreturn)
void ReportError(HWND window, LPCTSTR format, ...) {
	static array<TCHAR>* buffer = new array<TCHAR>();
	const size_t max_buffer_size = 4096;

	va_list args;
	HRESULT res;
	do {
		va_start(args, format);
		res = StringCchVPrintf(buffer->begin(), buffer->size(), format, args);
		va_end(args);

		if (res == STRSAFE_E_INSUFFICIENT_BUFFER && buffer->size() < max_buffer_size) {
			// May be able to correct this error by expanding the buffer.
			buffer->resize(2 * buffer->size());
			continue;
		}
		break;
	} while (1);

	MessageBox(window, buffer->begin(), _T("Error"), MB_ICONERROR | MB_OK);
	exit(EXIT_FAILURE);
}

LPCTSTR GetLastErrorString() {
	static array<TCHAR>* buffer = new array<TCHAR>();
	const size_t max_buffer_size = 4096;

	DWORD error = GetLastError();

	DWORD count;
	if (error == 0) {
		// Avoid the stereotypical "Error: Success" by recycling
		// the unknown error message for zero codes.
		count = 0;
	} else do {
		count = FormatMessage(
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			error,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			buffer->begin(), buffer->size(),
			NULL);

		if (count == 0) {
			if (GetLastError() == ERROR_INSUFFICIENT_BUFFER && buffer->size() < max_buffer_size) {
				// May be able to correct this error by expanding the buffer.
				buffer->resize(2 * buffer->size());
				continue;
			}
		}
		break;
	} while (1);

	if (count == 0) return _T("Unknown error");
	return buffer->begin();
}

static HRESULT CALLBACK DisplayWindowProc(HWND window, UINT code, WPARAM wparam, LPARAM lparam) {

	if (code == WM_PAINT) {
		PAINTSTRUCT info;
		if (BeginPaint(window, &info)) {
			DisplayPaint(window, info.hdc, info.fErase, info.rcPaint);
			EndPaint(window, &info);
		}
		return 0;
	}
	else if (code == WM_IDLE) {
		DisplayIdle(window);
		return 0;
	}
	else if (code == WM_CLOSE) {
		PostQuitMessage(0);
	}

	return DefWindowProc(window, code, wparam, lparam);
}

HWND CreateDisplayWindow(HINSTANCE instance, long width, long height) {

	DWORD style = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;

	RECT region = { 0, 0, width, height };
	if (!AdjustWindowRect(&region, style, FALSE)) {
		ReportError(NULL, _T("Failed to create display window: %s"), GetLastErrorString());
	}

	WNDCLASSEX window_class_config = {
		sizeof(window_class_config),
		CS_HREDRAW | CS_VREDRAW | CS_OWNDC,
		DisplayWindowProc,
		0, 0,
		instance,
		NULL,
		LoadCursor(NULL, IDC_ARROW),
		(HBRUSH)(COLOR_WINDOW+1),
		NULL, _T("EquisFtir Demo Window"), NULL
	};
	DWORD window_class = RegisterClassEx(&window_class_config);
	if (!window_class) {
		ReportError(NULL, _T("Failed to create display window: %s"), GetLastErrorString());
	}

	HWND window = CreateWindow(
		reinterpret_cast<LPCTSTR>(window_class),
		_T("EquisFtir Demo"),
		style,
		CW_USEDEFAULT, CW_USEDEFAULT,
		region.right - region.left,
		region.bottom - region.top,
		NULL,
		NULL,
		instance,
		NULL);
	if (!window) {
		ReportError(NULL, _T("Failed to create display window: %s"), GetLastErrorString());
	}

	return window;
}

int RunMessageLoop(HWND idle_window) {
	MSG msg;
	while (1) {
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) {
				return msg.wParam;
			}

			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		SendMessage(idle_window, WM_IDLE, 0, 0);
	}
}
