using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Threading;

namespace Demos.Rocks
{
    public class GuidedShape : Shape
    {
        private static int guidedShapeCount = 0;

        public delegate void ArrivedDelegate(GameTime time, GuidedShape shape);
        public event ArrivedDelegate Arrived;

        private Vector3 target;
        private float targetRadius;
        private float maxAcceleration;
        private float maxAngularVelocity;

        private float rotationTimeStep = 0;

        public GuidedShape(RocksGame game)
            : base(game)
        {
            Interlocked.Increment(ref guidedShapeCount);
        }

        public Vector3 Target
        {
            get
            {
                return target;
            }
            set
            {
                target = value;
            }
        }

        public float TargetRadius
        {
            get
            {
                return targetRadius;
            }
            set
            {
                targetRadius = value;
            }
        }

        public float MaxAcceleration
        {
            get
            {
                return maxAcceleration;
            }
            set
            {
                maxAcceleration = value;
            }
        }

        public float MaxAngularVelocity
        {
            get
            {
                return maxAngularVelocity;
            }
            set
            {
                maxAngularVelocity = value;
            }
        }

        private static Quaternion CreateFromVector(Vector3 vector)
        {
            Vector3 cross = Vector3.Cross(Vector3.UnitZ, vector);
            float dot = Vector3.Dot(Vector3.UnitZ, vector);
            if (cross == Vector3.Zero)
            {
                return Quaternion.Identity;
            }
            else
            {
                return Quaternion.Normalize(
                    new Quaternion(cross, vector.Length() + dot));
            }
        }

        public override void Update(GameTime gameTime)
        {
            Vector3 toTarget = Position - Target;

            bool arrived = false;
            if (Model != null)
            {
                foreach (ModelMesh mesh in Model.Meshes)
                {
                    arrived |= mesh.BoundingSphere.Contains(toTarget) != ContainmentType.Disjoint;
                }
            }

            // Have we reached the target? -- TODO account for passing through the target.
            if (arrived)
            {
                if (Arrived != null)
                    Arrived(gameTime, this);
                return;
            }

            rotationTimeStep += (float) gameTime.ElapsedGameTime.TotalSeconds;
            if (rotationTimeStep >= guidedShapeCount / 600.0f)
            {
                Quaternion rotationToTarget = CreateFromVector(Target - Position);

                float angle = (float) Math.Acos(Quaternion.Dot(Rotation, rotationToTarget)) * 2.0f;
                angle = Math.Min(angle, 2.0f * (float) Math.PI - angle);

                if (angle > 0.0f)
                {
                    float maxTurn = MaxAngularVelocity * rotationTimeStep;

                    // Proportion of the angle that could have been turned during the last step
                    float amount = Math.Min(1.0f, maxTurn / angle);

                    Rotation = Quaternion.Slerp(Rotation, rotationToTarget, amount);
                }

                float accel = 0.0f;
                if (angle < MathHelper.ToRadians(90.0f))
                {
                    accel = (float) Math.Cos(angle);
                    accel *= accel;
                    accel *= MaxAcceleration * rotationTimeStep;
                }

                Velocity = Vector3.Transform(Vector3.UnitZ, Rotation) * (Velocity.Length() + accel);

                rotationTimeStep = 0.0f;
            }

            base.Update(gameTime);
        }

        public override void Dispose()
        {
            base.Dispose();
            Interlocked.Decrement(ref guidedShapeCount);
        }
    }
}
