using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;

namespace Demos.Rocks
{
    public class Particles : Shape
    {
        private SpriteBatch batch;
        private Texture2D texture;
        private Vector2 textureSize;
        private Vector2 drift;

        private Texture2D font;
        private List<Particle> particles;

        public Particles(RocksGame game)
            : base(game)
        {
            this.batch = null;
            this.particles = new List<Particle>();
        }

        public Vector2 Drift
        {
            get
            {
                return drift;
            }
            set
            {
                drift = value;
            }
        }

        public void AddFtirEvent(Vector2 position, float mass, Vector4 color)
        {
            Particle p = new Particle();
            p.Position = position;

            p.StartColor = color;
            p.EndColor = new Vector4(color.X, color.Y, color.Z, 0.0f);

            p.StartScale = (float)Math.Sqrt(Math.Abs(mass)) / 128.0f;
            p.EndScale = p.StartScale / 4;
            p.Lifetime = 1.0f;
            particles.Add(p);
        }

        public void AddExplosion(Vector3 position, float radius)
        {
            // Project the position into screen space
            Vector3 screen3d = Game.GraphicsManager.GraphicsDevice.Viewport.Project(
                position, Game.Projection, Game.View, Matrix.Identity);

            Vector2 screen = new Vector2(screen3d.X, screen3d.Y);
            float maxOffset = radius * 0.5f;
            float maxVelocity = radius - maxOffset;

            for (int i = 0; i < 100; i++)
            {
                Vector2 offset = GameTools.RandomVector2(0.0f, maxOffset);
                Vector2 velocity = GameTools.RandomVector2(0.0f, maxVelocity);

                Particle particle = new Particle();
                particle.Position = screen + offset;
                particle.Velocity = velocity;
                particle.StartColor = new Vector4(1.0f, 1.0f - offset.Length() / maxOffset, 0.0f, 0.5f);
                particle.EndColor = new Vector4(0.25f, 0.25f, 0.25f, 0.0f);
                particle.StartScale = radius * 0.015f;
                particle.EndScale = radius * 0.04f;
                particle.Lifetime = GameTools.RandomFloat(0.5f, 1.0f);

                particles.Add(particle);
            }
        }

        public void AddDust(Vector3 position, float radius)
        {
            // Project the position into screen space
            Vector3 screen3d = Game.GraphicsManager.GraphicsDevice.Viewport.Project(
                position, Game.Projection, Game.View, Matrix.Identity);

            Vector2 screen = new Vector2(screen3d.X, screen3d.Y);
            float maxOffset = radius * 0.2f;
            float maxVelocity = radius - maxOffset;

            for (int i = 0; i < 50; i++)
            {
                Vector2 offset = GameTools.RandomVector2(0.0f, maxOffset);
                Vector2 velocity = GameTools.RandomVector2(0.0f, maxVelocity);

                Particle particle = new Particle();
                particle.Position = screen + offset;
                particle.Velocity = velocity; //139 90 43
                particle.StartColor = new Vector4(0.8f, 0.6f, 0.4f, 1.0f);
                particle.EndColor = new Vector4(0.4f, 0.3f, 0.2f, 0.0f);
                particle.StartScale = radius * 0.015f;
                particle.EndScale = radius * 0.04f;
                particle.Lifetime = GameTools.RandomFloat(0.5f, 1.0f);

                particles.Add(particle);
            }
        }

        public Particle AddText(String text, Vector3 position)
        {
            // Project the position into screen space
            Vector3 screen3d = Game.GraphicsManager.GraphicsDevice.Viewport.Project(
                position, Game.Projection, Game.View, Matrix.Identity);

            Vector2 screen = new Vector2(screen3d.X, screen3d.Y);
            return AddText(text, screen);
        }

        public Particle AddText(String text, Vector2 screenPosition)
        {
            Particle particle = new Particle();
            particle.Text = text;
            particle.Position = screenPosition;
            particle.StartColor = Vector4.One;
            particle.EndColor = Vector4.One;
            particle.StartScale = 1;
            particle.EndScale = 1;
            particle.Lifetime = float.PositiveInfinity;
            particles.Add(particle);

            return particle;
        }

        public override void Setup()
        {
            batch = new SpriteBatch(Game.GraphicsManager.GraphicsDevice);
            texture = Game.ContentManager.Load<Texture2D>(@"Textures\circle");
            textureSize = new Vector2(texture.Width, texture.Height);
            font = Game.ContentManager.Load<Texture2D>(@"Fonts\in-game_score");

            base.Setup();
        }

        public override void Update(GameTime gameTime)
        {
            float time = (float) gameTime.ElapsedGameTime.TotalSeconds;

            foreach (Particle particle in particles)
            {
                particle.Position += (particle.Velocity + drift)* time;
                particle.Ellapsed += time;
            }

            particles.RemoveAll(delegate(Particle particle)
            {
                return particle.Ellapsed > particle.Lifetime;
            });
        }

        public override void Draw(GameTime gameTime)
        {
            GraphicsDevice graphics = Game.GraphicsManager.GraphicsDevice;

            batch.Begin(SpriteBlendMode.AlphaBlend);
            graphics.RenderState.DepthBufferWriteEnable = false;

            foreach (Particle particle in particles)
            {
                Color color = new Color(Vector4.Lerp(
                    particle.StartColor, particle.EndColor,
                    particle.Ellapsed / particle.Lifetime));

                float scale = MathHelper.Lerp(
                    particle.StartScale, particle.EndScale,
                    particle.Ellapsed / particle.Lifetime);

                if (particle.Text == null)
                {
                    batch.Draw(
                        texture, particle.Position, null, color, 0.0f,
                        textureSize * particle.Origin,
                        scale,
                        SpriteEffects.None,
                        0.0f);
                }
                else
                {
                    int characterSpacing = 60;
                    int characterWidth = 58;
                    int characterHeight = 100;
                    Vector2 characterOrigin =
                        new Vector2(characterWidth, characterHeight)
                        * particle.Origin;

                    String text = particle.Text;
                    Vector2 textSize = new Vector2(
                        characterWidth * text.Length,
                        characterHeight) * scale;
                    Vector2 textPosition = particle.Position
                        - textSize * particle.Origin;
                    textPosition.Y += characterHeight * scale;

                    for (int i = 0; i < text.Length; ++i)
                    {
                        textPosition.X += characterWidth * scale;

                        Rectangle source = new Rectangle(
                            characterSpacing * (text[i] - '0'), 0,
                            characterSpacing, characterHeight);

                        batch.Draw(
                            font, textPosition, source, color, 0.0f,
                            characterOrigin,
                            scale,
                            SpriteEffects.None,
                            0.0f);
                    }
                }
            }

            graphics.RenderState.DepthBufferWriteEnable = true;
            batch.End();

            base.Draw(gameTime);
        }
    }
}
