using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Demos.Rocks
{
    public class Shape : IDisposable
    {
        public delegate void ExpiredDelegate(Shape shape);
        public event ExpiredDelegate Expired;

        private readonly RocksGame game;
        private Vector3 position = Vector3.Zero;
        private Vector3 velocity = Vector3.Zero;
        private Vector3 acceleration = Vector3.Zero;
        private Quaternion rotation = Quaternion.Identity;
        private Quaternion angularVelocity = Quaternion.Identity;
        private Vector3 scale = Vector3.One;
        private String modelPath = null;
        private Model model = null;
        private String texturePath = null;
        private Texture2D texture = null;
        private Vector3 ambientLightColor = Vector3.One;
        private Vector3 diffuseLightColor = Vector3.One;
        private Vector3 specularLightColor = Vector3.One;
        private float lifespan = float.PositiveInfinity;
        private bool deleted = false;

        public Shape(RocksGame game)
        {
            this.game = game;
        }

        public RocksGame Game
        {
            get
            {
                return game;
            }
        }

        public Vector3 Position
        {
            get
            {
                return position;
            }

            set
            {
                position = value;
            }
        }

        public Vector3 Velocity
        {
            get
            {
                return velocity;
            }
            set
            {
                velocity = value;
            }
        }

        public Vector3 Acceleration
        {
            get
            {
                return acceleration;
            }
            set
            {
                acceleration = value;
            }
        }

        public Quaternion Rotation
        {
            get
            {
                return rotation;
            }
            set
            {
                rotation = value;
            }
        }

        public Quaternion AngularVelocity
        {
            get
            {
                return angularVelocity;
            }
            set
            {
                angularVelocity = value;
            }
        }

        public Vector3 Scale
        {
            get
            {
                return scale;
            }
            set
            {
                scale = value;
            }
        }

        public String ModelPath
        {
            get
            {
                return modelPath;
            }

            set
            {
                modelPath = value;
            }
        }

        public Model Model
        {
            get
            {
                return model;
            }
        }

        public String TexturePath
        {
            get
            {
                return texturePath;
            }
            set
            {
                texturePath = value;
            }
        }

        public Texture2D Texture
        {
            get
            {
                return texture;
            }
        }

        public Vector3 AmbientLightColor
        {
            get
            {
                return ambientLightColor;
            }
            set
            {
                ambientLightColor = value;
            }
        }

        public Vector3 DiffuseLightColor
        {
            get
            {
                return diffuseLightColor;
            }
            set
            {
                diffuseLightColor = value;
            }
        }

        public Vector3 SpecularLightColor
        {
            get
            {
                return specularLightColor;
            }
            set
            {
                specularLightColor = value;
            }
        }

        public float Lifespan
        {
            get
            {
                return lifespan;
            }
            set
            {
                lifespan = value;
            }
        }

        public bool Deleted
        {
            get
            {
                return deleted;
            }
            set
            {
                deleted = value;
            }
        }

        public virtual void Setup()
        {
            if (texturePath != null)
            {
                texture = Game.ContentManager.Load<Texture2D>(TexturePath);
            }
            else
            {
                texture = null;
            }

            if (modelPath != null)
            {
                model = game.ContentManager.Load<Model>(modelPath);

                int meshIndex = 0;
                foreach (ModelMesh mesh in model.Meshes)
                {
                    foreach (BasicEffect effect in mesh.Effects)
                    {
                        effect.EnableDefaultLighting();

                        effect.AmbientLightColor = ambientLightColor;
                        effect.DiffuseColor = diffuseLightColor;
                        effect.SpecularColor = specularLightColor;

                        if (texture != null)
                        {
                            effect.TextureEnabled = true;
                            effect.Texture = texture;
                        }
                    }

                    meshIndex += 1;
                }
            }
            else
            {
                model = null;
            }
        }

        public virtual void Update(GameTime gameTime)
        {
            float time = (float) gameTime.ElapsedGameTime.TotalSeconds;

            if (lifespan == -1)
            {
                // Infinite lifespan
            }
            else if (time > lifespan)
            {
                Lifespan = 0.0f;
                Deleted = true;
                if (Expired != null)
                {
                    Expired(this);
                }
            }
            else
            {
                Lifespan -= time;
            }

            position += velocity * time;
            velocity += acceleration * time;

            rotation *= Quaternion.Slerp(Quaternion.Identity, angularVelocity, time);
        }

        public virtual void Draw(GameTime gameTime)
        {
            if (model != null)
            {
                Matrix world =
                    Matrix.CreateScale(scale) *
                    Matrix.CreateFromQuaternion(rotation) *
                    Matrix.CreateTranslation(position);

                GraphicsDevice graphics = Game.GraphicsManager.GraphicsDevice;
                graphics.RenderState.AlphaTestEnable = false;
                graphics.RenderState.AlphaBlendEnable = false;
                graphics.RenderState.DepthBufferEnable = true;
                graphics.RenderState.DepthBufferFunction = CompareFunction.LessEqual;
                graphics.RenderState.DepthBufferWriteEnable = true;

                foreach (ModelMesh mesh in model.Meshes)
                {
                    foreach (BasicEffect effect in mesh.Effects)
                    {
                        effect.World = mesh.ParentBone.Transform * world;
                        effect.Projection = game.Projection;
                        effect.View = game.View;
                    }

                    mesh.Draw();
                }
            }
        }

        public virtual void Dispose()
        {
            model = null;
        }
    }
}
