#include "common.hpp"
#include "TouchImpl.hpp"
#include "filter.hpp"
EQUIS_USING_NS1(equis);

ETSurface* etCreate() {
	ETSurface* handle = new ETSurface();

	handle->background_drift = 0.5f;
	handle->background_threshold = 5.0f;
	handle->pixel_threshold = 0.5f;
	handle->touch_threshold = 0.5f;

	handle->sticky_press = 2;
	handle->sticky_release = 5;

	return handle;
}

int etDelete(ETSurface* handle) {
	delete handle;
	return 0;
}

int etSourceVideoOpen(ETSurface* handle, const char* uri) {
	if (!handle) return -1;

	EVInput* video = eviOpen(uri);
	if (!video) return -1; // Must close video below this point.

	if (etSourceVideo(handle, video, 1) < 0) goto error;

	return 0;
error:
	eviClose(video);
	return -1;
}

static void processFrame(EVInputFrame* frame, void* surface) {
	static_cast<ETSurface*>(surface)->process(frame);
	evifDone(frame);
}

int etSourceVideo(ETSurface* handle, EVInput* video, int auto_close) {
	if (!handle) return -1;

	if (video) {
		EVFormat format = { EVF_TYPE_Y16 };
		if (eviSetFormat(video, &format) < 0) return -1;
	}

	if (handle->video) {
		eviCallback(handle->video, NULL, NULL);
		if (handle->close_video) eviClose(handle->video);
	}

	handle->video = video;
	handle->close_video = auto_close;

	eviCallback(video, processFrame, handle);

	return 0;
}

long etPoll(ETSurface* handle, ETEvent* events, long events_cap) {
	if (!handle || !events) return -1;
	long count = handle->track.Poll(events, events_cap);
	for (long i = 0; i < count; ++i) {
		vec2f v;
		v.x = events[i].x;
		v.y = events[i].y;
		v = handle->calib.transform(v);
		events[i].x = v.x;
		events[i].y = v.y;
	}
	return count;
}

long etNext(ETSurface* handle, ETEvent* events, long events_cap) {
	return -1;
}

int etCalibrate(ETSurface* handle, const float* desired, const float* reported, long end) {
	if (desired == NULL || reported == NULL || end == 0) {
		handle->calib.clear();
		return 0;
	}

	array<vec2f> desired_points((vec2f*)desired, end / 2);
	array<vec2f> reported_points((vec2f*)reported, end / 2);
	if (!handle->calib.update(desired_points, reported_points)) return -1;
	return 0;
}

int etGetFormat(ETSurface* handle, EVFormat* format) {
	if (!handle || !format) return -1;
	format->types = EVF_TYPE_A8R8G8B8;
	format->flags = 0;
	format->width = handle->filtered.cols();
	format->height = handle->filtered.rows();
	format->usecs = 0;
	return 0;
}

#include "common/image/pixel.hpp"
using equis::image::PixelA8R8G8B8;

long etPreview(ETSurface* handle, int source, void* image, long image_stride, long image_cap) {
	if (!handle || source < 0 || source > ET_PREVIEW_EVENTS || !image) return -1;

	array2<uint16>* source_image;
	switch (source) {
		case ET_PREVIEW_INPUT: source_image = &handle->converted; break;
		case ET_PREVIEW_FILTERED: source_image = &handle->filtered; break;
		default: source_image = NULL;
	}
	if (!source_image) return -1;

	long image_extent = sizeof(PixelA8R8G8B8) * source_image->size();
	if (image_cap < image_extent) return -1;

	for (long row = 0; row < source_image->rows(); ++row) {
		PixelA8R8G8B8* pixels = reinterpret_cast<PixelA8R8G8B8*>(
			byte_offset(image, image_stride * row));
		
		for (long col = 0; col < source_image->cols(); ++col) {
			pixels[col].a = 0xFF;
			pixels[col].r = pixels[col].g = pixels[col].b = source_image->at(row, col) >> 8;
		}
	}

	return image_extent;
}

#if 0
EVInputFrame* etPreview(ETSurface* handle, int source) {
	if (!handle || source < 0 || source > ET_PREVIEW_EVENTS) return NULL;

	void* data = NULL;
	switch (source) {
		case ET_PREVIEW_INPUT: data = handle->converted.begin(); break;
		case ET_PREVIEW_BACKGROUND: data = handle->background.begin(); break;
		//case ET_PREVIEW_FOREGROUND: data = handle->foreground.begin(); break;
		case ET_PREVIEW_FILTERED: data = handle->filtered.begin(); break;
		case ET_PREVIEW_BLOBS:
		case ET_PREVIEW_TOUCHES:
		case ET_PREVIEW_EVENTS: break;
	}
	if (!data) return NULL;

	return evifCreateView(&handle->process_format, data, evfSize(&handle->process_format));
}
#endif

int etGeti(ETSurface* handle, int name, int* value) {
	if (!handle || !value) return -1;

	switch (name) {
		case ET_NOISE_FILTER: *value = handle->noise_filter;
		case ET_STICKY_PRESS: *value = handle->sticky_press;
		case ET_STICKY_RELEASE: *value = handle->sticky_release;
		default: return -1;
	}

	return 0;
}

int etGetf(ETSurface* handle, int name, float* value) {
	if (!handle || !value) return -1;

	switch (name) {
		case ET_BACKGROUND_DRIFT: *value = handle->background_drift; break;
		case ET_BACKGROUND_THRESHOLD: *value = handle->background_threshold; break;
		case ET_PIXEL_THRESHOLD: *value = handle->pixel_threshold; break;
		case ET_TOUCH_THRESHOLD: *value = handle->touch_threshold; break;
		default: return -1;
	}

	return 0;
}

int etSeti(ETSurface* handle, int name, int value) {
	if (!handle) return -1;

	switch (name) {
		case ET_NOISE_FILTER: handle->noise_filter = value; break;
		case ET_STICKY_PRESS: handle->sticky_press = value; break;
		case ET_STICKY_RELEASE: handle->sticky_release = value; break;
		default: return -1;
	}

	return 0;
}

int etSetf(ETSurface* handle, int name, float value) {
	if (!handle) return -1;

	switch (name) {
		case ET_BACKGROUND_DRIFT: handle->background_drift = value; break;
		case ET_BACKGROUND_THRESHOLD: handle->background_threshold = value; break;
		case ET_PIXEL_THRESHOLD: handle->pixel_threshold = value; break;
		case ET_TOUCH_THRESHOLD: handle->touch_threshold = value; break;
		default: return -1;
	}

	return 0;
}


static void invert(array2<pixel_rgb32>& frame) {
	for (long row = 0; row < frame.rows(); ++row) {
		for (long col = 0; col < frame.cols(); ++col) {
			pixel_rgb32& pixel = frame(row, col);
			pixel.r = 255 - pixel.r;
			pixel.g = 255 - pixel.g;
			pixel.b = 255 - pixel.b;
		}
	}
}

static void background_create(array2<uint16>& bg, array2<uint16>& frame, long n) {
	for (long row = 0; row < frame.rows(); ++row) {
		for (long col = 0; col < frame.cols(); ++col) {
			bg(row, col) = min(bg(row, col), frame(row, col));
		}
	}
}

static void background_apply(array2<uint16>& bg, array2<uint16>& frame) {
	for (long row = 0; row < frame.rows(); ++row) {
		for (long col = 0; col < frame.cols(); ++col) {
			uint16 bp = bg(row, col);
			uint16& fp = frame(row, col);

			if (fp < bp) fp = 0;
			else fp -= bp;
		}
	}
}

static LARGE_INTEGER s_perf_total;

void ETSurface::process(EVInputFrame* input) {
	const EVFormat* format = evifFormat(input);
	if (format->types != EVF_TYPE_Y16) {
		_ASSERT(FALSE);
		return;
	}

#if 0
	if (frame_index > 300) {
		LARGE_INTEGER freq;
		QueryPerformanceFrequency(&freq);
		printf("Performance estimate: %0.2lf fps\n", 
			frame_index / (s_perf_total.QuadPart / (double)freq.QuadPart));
		fgetc(stdin);
		exit(0);
	}
#endif

	const uint16* data = reinterpret_cast<const uint16*>(evifPointer(input));

	const int background_end = background_start + background_length;

	const int outer = 16;
	const int inner = 8;

	if (process_format.width != format->width || process_format.height != format->height) {
		process_format.types = EVF_TYPE_Y16;
		process_format.flags = 0;
		process_format.width = format->width;
		process_format.height = format->height;
		process_format.usecs = format->usecs;

		detect.Stats().SetWindow(4);

		track.SetMinAge(sticky_press);
		track.SetMaxAge(sticky_release);
		track.Resize(format->height, format->width);

		converted.resize(format->height, format->width);
		temp.resize(format->height, format->width);
		filtered.resize(format->height, format->width);
		marked.resize(format->height, format->width);
		background.resize(format->height, format->width, 0xFFFF);
	}

	detect.SetThreshold(static_cast<uint16>(pixel_threshold * 1e4f));
	detect.SetMinStrength(touch_threshold * 1e4f);

	frame_index += 1;
	
	if (background_mode > 0 && frame_index < background_start) {
		// Skip frames until the background starts.
		return;
	}

	/*if (config.invert) {
		invert(frame);
	}*/

	copy(data, byte_offset(data, evifSize(input)), converted.begin());
	/*switch (config.convert) {
			case 0: image::convert_mean(frame, converted); break;
			case 1: image::convert_lum(frame, converted); break;
			case 2: image::convert_bias(frame, converted); break;
			default: EQUIS_ASSERT(false);
	}*/

	if (background_mode == 1) {
		if (frame_index < background_end) {
			background_create(background, converted, background_length);
		} else {
			background_apply(background, converted);
		}
	}

	LARGE_INTEGER start;
	QueryPerformanceCounter(&start);

	switch (noise_mode) {
			case 0: copy(converted.begin(), converted.end(), filtered.begin()); break;
			case 1: band_low(converted, temp, filtered, inner); break;
			case 2: band_mid(converted, temp, filtered, outer, inner); break;
			default: _ASSERT(false);
	}

	LARGE_INTEGER end;
	QueryPerformanceCounter(&end);
	s_perf_total.QuadPart += end.QuadPart - start.QuadPart;

	if (background_mode == 2) {
		if (frame_index < background_end) {
			background_create(background, filtered, background_length);
		} else {
			background_apply(background, filtered);
		}
	}

	if (background_mode > 0 && frame_index < background_end) {
		// Skip event processing until the background is defined.
		return;
	}

	detect.Update(filtered);

	vector<Blob> blobs;
	detect.Peek(blobs);

	double trigger = touch_threshold * 1e4;
	FrameStats& stats = detect.Stats();

	for (size_t i = 0; i < blobs.size(); ++i) {
		blobs[i].mass -= blobs[i].area * detect.GetThreshold();
	}

#if 0
	/*if (config.verbose) {*/
		fprintf(stderr, "%10.2lf ", trigger / 1e4);
		for (size_t i = 0; i < 3; ++i) {
			if (blobs.size() > i) {
				float mass = blobs[i].mass;
				if (mass > trigger) {
					fprintf(stderr, " %10.2f!", mass / 1e4);
				} else {
					fprintf(stderr, " %10.2f ", mass / 1e4);
				}
			}
		}
		fputc('\n', stderr);
	/*}*/
#endif

	for (size_t i = 0; i < blobs.size(); ++i) {
		if (blobs[i].mass >= trigger) {
			stats.RemoveSample(blobs[i].mass);
		} else {
			blobs.erase(blobs.begin() + i, blobs.end());
			break;
		}
	}

	track.Update(blobs);

	/*if (config.output) {
		for (size_t i = 0; i < events.size(); ++i) {
			image::BlobTracker::Event& event(events[i]);

			const char* op;
			if (event.type > 0) op = "add";
			else if (event.type < 0) op = "del";
			else op = "set";

			fprintf(config.output, "%ld %s %ld %ld",
				frame_index, op, event.blob.cursor_id, event.blob.session_id);
			if (event.type >= 0) {
				fprintf(config.output, " %f %f",
					event.blob.x / frame.cols(),
					event.blob.y / frame.rows());
			}
			fputc('\n', config.output);

			fflush(config.output);
		}
	}*/

	/*if (config.preview) {
		uint16 mark = detect.GetThreshold();
		for (size_t row = 0; row < filtered.rows(); ++row) {
			for (size_t col = 0; col < filtered.cols(); ++col) {
				if (filtered(row, col) >= mark) {
					marked(row, col) = static_cast<uint16>(min(65535L, 2L * filtered(row,col)));
				} else {
					marked(row, col) = 0;
					filtered(row, col) /= 2;
				}
			}
		}

		if (config.mark_blobs & 1) {
			image::combine(filtered, marked, filtered, frame);
		} else {
			image::combine(filtered, filtered, filtered, frame);
		}
		if (config.mark_blobs & 2) {
			// TODO: Probably ought to use the blob's actual bounds.

		}
		if (fwrite(frame.begin(), frame.byte_size(), 1, config.preview) < 1) {
			break;
		}
		fflush(config.preview);
	}*/
}
