#include "common.hpp"
#include "sse.hpp"
#include "filter.hpp"

// Width of a column slice in uint16s.
#define COL_SLICE_WIDTH (128/16)

namespace {
	struct BandLow {
		int outer;
		int inner;
	
		int rows;
		int cols;
	
		int shift;
	
		void apply_row(const uint16* src, uint16* dst);
		void apply_col(const uint16* src, uint16* dst);
		void apply_col_slice(const uint16* src, uint16* dst);
	};
}

void
band_low(const array2<uint16>& src, array2<uint16>& tmp, array2<uint16>& dst, int outer) {
	
	BandLow s;
	s.outer = outer;
	s.inner = outer / 2;
	s.rows = src.rows();
	s.cols = src.cols();

	s.shift = static_cast<int>(log(0.25 * outer * outer * outer) / log(2.0));

	tmp.resize(s.rows, s.cols);
	dst.resize(s.rows, s.cols);

	// Process from img into tmp
	for (int r = 0; r < s.rows; ++r) {
		s.apply_row(src.row(r), tmp.row(r));
	}

	// Process from tmp back to image
	for (int c = 0; c < s.cols; ++c) {
		s.apply_col(tmp.begin() + c, dst.begin() + c);
	}
}

void BandLow::
apply_row(const uint16* src, uint16* dst) {
	
	long a0 = 0;
	long a1 = 0;
	long a2 = 0;

	for (int c = -outer; c < 0; ++c) {
		a0 += a1; a1 += a2;
		a2 += 1L * src[c + outer];
		if (c >= -inner) a2 -= 2L * src[c + inner];
	}

	for (int c = 0; c < cols; ++c) {
		dst[c] = static_cast<uint16>(a0 >> shift);
		a0 += a1; a1 += a2;
		if (c < cols - outer) a2 += 1L * src[c + outer];
		if (c < cols - inner) a2 -= 2L * src[c + inner];
		if (c >= inner) a2 += 2L * src[c - inner];
		if (c >= outer) a2 -= 1L * src[c - outer];
	}
}

void BandLow::
apply_col(const uint16* src, uint16* dst) {
	long a0 = 0;
	long a1 = 0;
	long a2 = 0;

	for (int r = -outer; r < 0; ++r) {
		a0 += a1; a1 += a2;

		if (r >= -outer) a2 += 1L * src[(r + outer) * cols];
		if (r >= -outer/2) a2 -= 2L * src[(r + inner) * cols];
	}

	for (int r = 0; r < rows; ++r) {
		dst[r*cols] = static_cast<uint16>(a0 >> shift);

		a0 += a1; a1 += a2;

		if (r < rows - outer) a2 += 1L * src[(r + outer) * cols];
		if (r < rows - outer/2) a2 -= 2L * src[(r + outer/2) * cols];
		if (r >= outer/2) a2 += 2L * src[(r - outer/2) * cols];
		if (r >= outer) a2 -= 1L * src[(r - outer) * cols];
	}
}

void BandLow::
apply_col_slice(const uint16* src, uint16* dst) {

	// TODO: Check stream vs store performance. Will usually be using the output
	// data immediately, so the store might be faster if everything fits in cache.

	// TODO: Repeatedly adding to a?2 might limit parallel ops. Consider manually
	// splitting the adds across some temporary registers.

	const __m128i zero = _mm_set1_epi32(0);

	__m128i aH0 = _mm_set1_epi32(0);
	__m128i aH1 = _mm_set1_epi32(0);
	__m128i aH2 = _mm_set1_epi32(0);

	__m128i aL0 = _mm_set1_epi32(0);
	__m128i aL1 = _mm_set1_epi32(0);
	__m128i aL2 = _mm_set1_epi32(0);

	for (int r = -outer; r < rows; ++r) {
		if (r >= 0) {
			_mm_stream_si128(dst + r * cols, _mm_packs_epi32(
				_mm_srli_epi32(aL0, shift),
				_mm_srli_epi32(aH0, shift)));
		}

		aH0 = _mm_add_epi32(aH0, aH1);
		aL0 = _mm_add_epi32(aL0, aL1);

		aH1 = _mm_add_epi32(aH1, aH2);
		aL1 = _mm_add_epi32(aL1, aL2);

		if (r >= -outer && r < rows - outer) {
			__m128i in = _mm_load_si128(src + (r + outer) * cols);
			aH2 = _mm_add_epi32(aH2, _mm_unpackhi_epi16(in, zero));
			aL2 = _mm_add_epi32(aL2, _mm_unpacklo_epi16(in, zero));
		}

		if (r >= -inner && r < rows - inner) {
			__m128i in = _mm_load_si128(src + (r + inner) * cols);
			aH2 = _mm_sub_epi32(aH2, _mm_slli_epi32(_mm_unpackhi_epi16(in, zero), 1));
			aL2 = _mm_sub_epi32(aL2, _mm_slli_epi32(_mm_unpacklo_epi16(in, zero), 1));
		}

		if (r >= +inner) {
			__m128i in = _mm_load_si128(src + (r - inner) * cols);
			aH2 = _mm_add_epi32(aH2, _mm_slli_epi32(_mm_unpackhi_epi16(in, zero), 1));
			aL2 = _mm_add_epi32(aL2, _mm_slli_epi32(_mm_unpacklo_epi16(in, zero), 1));
		}
		
		if (r >= +outer) {
			__m128i in = _mm_load_si128(src + (r - outer) * cols);
			aH2 = _mm_sub_epi32(aH2, _mm_unpackhi_epi16(in, zero));
			aL2 = _mm_sub_epi32(aL2, _mm_unpacklo_epi16(in, zero));
		}
	}
}
