#include "common.hpp"
#include "sse.hpp"
#include "filter.hpp"

namespace {
	struct BandMid {
		int outer;
		int inner;
	
		int rows;
		int cols;
	
		int shift;
		int inner_shift;
	
		void apply_row(const uint16* src, uint16* dst);
		void apply_col(const uint16* src, uint16* dst);
	};

	static const uint16 zeros[64];
}

void
band_mid(const array2<uint16>& src, array2<uint16>& tmp, array2<uint16>& dst, int outer, int inner) {
	
	BandMid s;
	s.outer = outer;
	s.inner = inner;

	s.rows = src.rows();
	s.cols = src.cols();

	// outer = 16, inner = 8, shift = 3, add 1
	// outer = 16, inner = 4, shift = 6, add 4
	s.inner_shift = 3;
	s.shift = static_cast<int>(log(0.25 * inner * inner * inner) / log(2.0)) + 1;

	tmp.resize(s.rows, s.cols);
	dst.resize(s.rows, s.cols);

	// Process from img into tmp
	for (int r = 0; r < s.rows; ++r) {
		s.apply_row(src.row(r), tmp.row(r));
	}

	// Process from tmp back to image
	for (int c = 0; c < s.cols; c += 1) {
		s.apply_col(tmp.begin() + c, dst.begin() + c);
	}

	// Clear the corners to avoid artifacts
	for (int r = 0; r < s.outer; ++r) {
		for (int c = 0; c < s.outer; ++c) {
			dst(r, c) = 0;
			dst(dst.rows() - 1 - r, c) = 0;
			dst(r, dst.cols() - 1 - c) = 0;
			dst(dst.rows() - 1 - r, dst.cols() - 1 - c) = 0;
		}
	}
}

void BandMid::
apply_row(const uint16* src, uint16* dst) {
	
	const int outer = 16;
	const int inner = 8;
	const int inner_shift = 0;
	const int shift = 8;

	const int im = 8;

	long a0 = 0;
	long a1 = 0;
	long a2 = 0;

	for (int c = -outer; c < 0; ++c) {
		a0 += a1; a1 += a2;

		if (c >= -outer) a2 -= 1L * src[c + outer];
		if (c >= -outer/2) a2 += 2L * src[c + outer/2];

		if (c >= -inner) a2 += 1L * im * src[c + inner];
		if (c >= -inner/2) a2 -= 2L * im * src[c + inner/2];
	}

	for (int c = 0; c < outer; ++c) {
		dst[c] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		if (c < cols - outer) a2 -= 1L * src[c + outer];
		if (c < cols - outer/2) a2 += 2L * src[c + outer/2];
		if (c >= outer/2) a2 -= 2L * src[c - outer/2];
		if (c >= outer) a2 += 1L * src[c - outer];

		if (c < cols - inner) a2 += 1L * im * src[c + inner];
		if (c < cols - inner/2) a2 -= 2L * im * src[c + inner/2];
		if (c >= inner/2) a2 += 2L * im * src[c - inner/2];
		if (c >= inner) a2 -= 1L * im * src[c - inner];
	}

	for (int c = outer; c < cols - outer; ++c) {
		dst[c] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		a2 -= 1L * src[c + outer];
		a2 += 2L * src[c + outer/2];
		a2 -= 2L * src[c - outer/2];
		a2 += 1L * src[c - outer];

		a2 += 1L * im * src[c + inner];
		a2 -= 2L * im * src[c + inner/2];
		a2 += 2L * im * src[c - inner/2];
		a2 -= 1L * im * src[c - inner];
	}

	for (int c = cols - outer; c < cols; ++c) {
		dst[c] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		if (c < cols - outer) a2 -= 1L * src[c + outer];
		if (c < cols - outer/2) a2 += 2L * src[c + outer/2];
		if (c >= outer/2) a2 -= 2L * src[c - outer/2];
		if (c >= outer) a2 += 1L * src[c - outer];

		if (c < cols - inner) a2 += 1L * im * src[c + inner];
		if (c < cols - inner/2) a2 -= 2L * im * src[c + inner/2];
		if (c >= inner/2) a2 += 2L * im * src[c - inner/2];
		if (c >= inner) a2 -= 1L * im * src[c - inner];
	}
}

void BandMid::
apply_col(const uint16* src, uint16* dst) {

	const int outer = 16;
	const int inner = 8;
	const int inner_shift = 0;
	const int shift = 8;

	const int im = 8;

	long a0 = 0;
	long a1 = 0;
	long a2 = 0;

	//_ASSERT((DWORD_PTR)src % 16 == 0);
	//_ASSERT((DWORD_PTR)dst % 16 == 0);

	for (int r = -outer; r < 0; ++r) {
		a0 += a1; a1 += a2;

		if (r >= -outer) a2 -= 1L * src[(r + outer) * cols];
		if (r >= -outer/2) a2 += 2L * src[(r + outer/2) * cols];

		if (r >= -inner) a2 += 1L*im * src[(r + inner) * cols];
		if (r >= -inner/2) a2 -= 2L*im * src[(r + inner/2) * cols];
	}

	for (int r = 0; r < outer; ++r) {
		dst[r*cols] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		if (r < rows - outer) a2 -= 1L * src[(r + outer) * cols];
		if (r < rows - outer/2) a2 += 2L * src[(r + outer/2) * cols];
		if (r >= outer/2) a2 -= 2L * src[(r - outer/2) * cols];
		if (r >= outer) a2 += 1L * src[(r - outer) * cols];

		if (r < rows - inner) a2 += 1L*im * src[(r + inner) * cols];
		if (r < rows - inner/2) a2 -= 2L*im * src[(r + inner/2) * cols];
		if (r >= inner/2) a2 += 2L*im * src[(r - inner/2) * cols];
		if (r >= inner) a2 -= 1L*im * src[(r - inner) * cols];
	}

	for (int r = outer; r < rows - outer; ++r) {
		dst[r*cols] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		a2 -= 1L * src[(r + outer) * cols];
		a2 += 2L * src[(r + outer/2) * cols];
		a2 -= 2L * src[(r - outer/2) * cols];
		a2 += 1L * src[(r - outer) * cols];

		a2 += 1L * im * src[(r + inner) * cols];
		a2 -= 2L * im * src[(r + inner/2) * cols];
		a2 += 2L * im * src[(r - inner/2) * cols];
		a2 -= 1L * im * src[(r - inner) * cols];
	}

	for (int r = rows - outer; r < rows; ++r) {
		dst[r*cols] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		if (r < rows - outer) a2 -= 1L * src[(r + outer) * cols];
		if (r < rows - outer/2) a2 += 2L * src[(r + outer/2) * cols];
		if (r >= outer/2) a2 -= 2L * src[(r - outer/2) * cols];
		if (r >= outer) a2 += 1L * src[(r - outer) * cols];

		if (r < rows - inner) a2 += 1L*im * src[(r + inner) * cols];
		if (r < rows - inner/2) a2 -= 2L*im * src[(r + inner/2) * cols];
		if (r >= inner/2) a2 += 2L*im * src[(r - inner/2) * cols];
		if (r >= inner) a2 -= 1L*im * src[(r - inner) * cols];
	}
}

#if 0
void BandMid::
apply_col(const uint16* src, uint16* dst) {
	const int outer = 16;
	const int inner = 8;
	const int inner_shift = 0;
	const int shift = 8;

	const int im = 8;

	long a0 = 0;
	long a1 = 0;
	long a2 = 0;

	for (int r = -outer; r < 0; ++r) {
		a0 += a1; a1 += a2;

		if (r >= -outer) a2 -= 1L * src[(r + outer) * cols];
		if (r >= -outer/2) a2 += 2L * src[(r + outer/2) * cols];

		if (r >= -inner) a2 += 1L*im * src[(r + inner) * cols];
		if (r >= -inner/2) a2 -= 2L*im * src[(r + inner/2) * cols];
	}

	for (int r = 0; r < outer; ++r) {
		dst[r*cols] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		if (r < rows - outer) a2 -= 1L * src[(r + outer) * cols];
		if (r < rows - outer/2) a2 += 2L * src[(r + outer/2) * cols];
		if (r >= outer/2) a2 -= 2L * src[(r - outer/2) * cols];
		if (r >= outer) a2 += 1L * src[(r - outer) * cols];

		if (r < rows - inner) a2 += 1L*im * src[(r + inner) * cols];
		if (r < rows - inner/2) a2 -= 2L*im * src[(r + inner/2) * cols];
		if (r >= inner/2) a2 += 2L*im * src[(r - inner/2) * cols];
		if (r >= inner) a2 -= 1L*im * src[(r - inner) * cols];
	}

	for (int r = outer; r < rows - outer; ++r) {
		dst[r*cols] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		a2 -= 1L * src[(r + outer) * cols];
		a2 += 2L * src[(r + outer/2) * cols];
		a2 -= 2L * src[(r - outer/2) * cols];
		a2 += 1L * src[(r - outer) * cols];

		a2 += 1L * im * src[(r + inner) * cols];
		a2 -= 2L * im * src[(r + inner/2) * cols];
		a2 += 2L * im * src[(r - inner/2) * cols];
		a2 -= 1L * im * src[(r - inner) * cols];
	}

	for (int r = rows - outer; r < rows; ++r) {
		dst[r*cols] = static_cast<uint16>(clamp(a0 >> shift, 0L, 65535L));

		a0 += a1; a1 += a2;

		if (r < rows - outer) a2 -= 1L * src[(r + outer) * cols];
		if (r < rows - outer/2) a2 += 2L * src[(r + outer/2) * cols];
		if (r >= outer/2) a2 -= 2L * src[(r - outer/2) * cols];
		if (r >= outer) a2 += 1L * src[(r - outer) * cols];

		if (r < rows - inner) a2 += 1L*im * src[(r + inner) * cols];
		if (r < rows - inner/2) a2 -= 2L*im * src[(r + inner/2) * cols];
		if (r >= inner/2) a2 += 2L*im * src[(r - inner/2) * cols];
		if (r >= inner) a2 -= 1L*im * src[(r - inner) * cols];
	}
}
#endif