#include "common.hpp"
#include "stats.hpp"

//EQUIS_USING_NS1(equis);

/*
 * Incremental mean:
 * A[1] = x[1]
 * A[i] = A[i-1] + (x[i] - A[i-1]) / i
 *
 * Incremental standard deviation (partial):
 * Q[1] = 0
 * Q[i] = Q[i-1] + (x[i] - A[i-1]) * (x[i] - A[i])
 *
 * Removing one sample:
 * A[i-1] = (A[i] * i - x[i]) / (i - 1)
 * Q[i-1] = Q[i] - (x[i] - A[i-1]) * (x[i] - A[i])
 *
 * Removing many samples: TODO, derive fast stddev
 * A[i-N] = (A[i] * i - Sum) / (i - N)
 *
 */

FrameStats::
FrameStats()
    : m_window(1), m_count(0), m_mean(0.0f), m_dev(0.0f), m_curr(NULL) {
}

FrameStats::
~FrameStats() {
}

void FrameStats::
SetWindow(size_t frames) {
    if (frames < 1) frames = 1;
    m_window = frames;
}

size_t FrameStats::
GetWindow() const {
    return m_window;
}

void FrameStats::
EnterFrame() {
   
    /* Remove expired frames. */
    while (m_past.size() >= m_window) {
        Frame& past = m_past.front();

        if (m_count <= past.count) {
            m_count = 0;
            m_mean = 0.0f;
            m_dev = 0.0f;

        } else {
            const long old_count = m_count;

            m_count -= past.count;
            m_mean = (m_mean * old_count - past.sum) / m_count;
            m_dev -= past.dev;
        }

        m_past.pop();
    }

    /* Add a new empty frame. */
    m_past.push(Frame());
    m_curr = &m_past.back();
}

void FrameStats::
AddSample(double value) {
    const double old_mean = m_mean;
    m_count += 1;
    m_mean += (value - m_mean) / m_count;

    const double dev_step = (value - old_mean) * (value - m_mean);
    m_dev += dev_step;
    
    if (m_curr) {
        m_curr->count += 1;
        m_curr->sum += value;
        m_curr->dev  += dev_step;
    }
}

void FrameStats::
LeaveFrame() {
}

void FrameStats::
RemoveSample(double value) {
    
    const double old_mean = m_mean;
    m_mean = (m_mean * m_count - value) / (m_count - 1);

    const double dev_step = (value - old_mean) * (value - m_mean);
    m_dev -= dev_step;

    m_count -= 1;

    if (m_curr) {
        m_curr->count -= 1;
        m_curr->sum -= value;
        m_curr->dev -= dev_step;
    }
}

long FrameStats::
GetCount() const {
    return m_count;
}

double FrameStats::
GetMean() const {
    return m_mean;
}

double FrameStats::
GetStdDev() const {
    return sqrt(m_dev / m_count);
}

void FrameStats::
WithoutSample(double value, double& mean, double& stddev) {

    mean = (m_mean * m_count - value) / (m_count - 1);

	const double dev_step = (value - m_mean) * (value - mean);
	stddev = sqrt((m_dev - dev_step) / (m_count - 1));
}

FrameStats::Frame::
Frame()
: count(0), sum(0.0f), dev(0.0f) {
}
