#include "common.hpp"
#include "common/image/image.hpp"
#include "DSInput.hpp"
#include "DSInputFrame.hpp"
EQUIS_USING_NS2(equis,image);

#include <emmintrin.h>

static HRESULT ConvertSample(
	const EVFormat* dfmt, void* dst,
	const AM_MEDIA_TYPE* sample_type, IMediaSample* sample);

DSInputFrame::DSInputFrame(DSInput* origin)
: next(NULL)
{
	this->origin = origin;
}

DSInputFrame::~DSInputFrame() {
}

HRESULT DSInputFrame::SetFormat(const EVFormat* format) {
	HRESULT res = S_OK;

	if (format == NULL) res = E_POINTER;

	m_buffer.resize(evfSize(format));

	if (res == S_OK) {
		m_format = *format;
	}

	_ASSERT(res == S_OK);
	return res;
}

static bool FormatsDiffer(const EVFormat& a, const EVFormat& b) {
	return
		a.types != b.types ||
		a.flags != b.flags ||
		a.width != b.width ||
		a.height != b.height ||
		a.usecs != b.usecs;
}

HRESULT DSInputFrame::SetSample(const AM_MEDIA_TYPE* type, IMediaSample* sample) {

	m_sample.Detach();
	if (sample == NULL) return S_OK;

	HRESULT res = S_OK;

	if (res == S_OK) {
		res = ConvertSample(&m_format, m_buffer.begin(), type, sample);
		if (res == S_FALSE) {
			// No conversion necessary; use the sample buffer.
			m_sample.Attach(sample);
			res = S_OK;
		}
	}

	_ASSERT(res == S_OK);
	return res;
}

const EVFormat* DSInputFrame::Format() const {
	return &m_format;
}

long DSInputFrame::Size() const {
	if (m_sample.IsNull()) {
		return m_buffer.size();
	}

	return m_sample->GetActualDataLength();
}

void* DSInputFrame::Pointer() {
	if (m_sample.IsNull()) {
		return m_buffer.begin();
	}

	BYTE* ptr = NULL;
	if (m_sample->GetPointer(&ptr) != S_OK) return NULL;
	return ptr;
}

static HRESULT ConvertSample(
	const EVFormat* dfmt, void* dst,
	const AM_MEDIA_TYPE* sample_type, IMediaSample* sample)
{
	HRESULT res = S_OK;

	if (!IsEqualGUID(sample_type->majortype, WMMEDIATYPE_Video) ||
		!IsEqualGUID(sample_type->formattype, WMFORMAT_VideoInfo) ||
		!sample_type->bFixedSizeSamples ||
		sample_type->bTemporalCompression)
	{
		res = E_INVALIDARG;
	}

	// TODO: Sanity check the actual data length.
	BYTE* sample_data;
	if (res == S_OK) {
		res = sample->GetPointer(&sample_data);
	}

	long rows = dfmt->height;
	long cols = dfmt->width;
	bool vflip = false;

	if (res != S_OK) {} // Skip conversions
	else if (dfmt->types == EVF_TYPE_A8R8G8B8) {
		if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_RGB32)) {
			res = S_FALSE; // No conversion necessary.
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_RGB24)) {
			convertImage<A8R8G8B8, R8G8B8>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_Y800)) {
			convertImage<A8R8G8B8, Y8>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_Y160)) {
			convertImage<A8R8G8B8, Y16>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_I420)) {
			convertImage<A8R8G8B8, I420>(dst, sample_data, rows, cols, vflip);
		}
		else {
			// Unable to convert this format to RGB32.
			res = E_FAIL;
		}
	}
	else if (dfmt->types == EVF_TYPE_Y8)  {
		if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_RGB32)) {
			convertImage<Y8, A8R8G8B8>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_RGB24)) {
			convertImage<Y8, R8G8B8>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_Y800) ||
				IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_I420))
		{
			// I420 starts with an 8-bpp luminance plane,
			// so no conversion is actually necessary.
			res = S_FALSE; // No conversion necessary.
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_Y160)) {
			convertImage<Y8, Y16>(dst, sample_data, rows, cols, vflip);
		}
		else {
			// Unable to convert this format to Y8.
			res = E_FAIL;
		}
	}
	else if (dfmt->types == EVF_TYPE_Y16) {
		if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_RGB32)) {
			convertImage<Y16, A8R8G8B8>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_RGB24)) {
			convertImage<Y16, R8G8B8>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_Y800) ||
				 IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_I420))
		{
			// I420 starts with an 8-bpp luminance plane,
			// so can reuse the conversion from Y8.
			convertImage<Y16, Y8>(dst, sample_data, rows, cols, vflip);
		}
		else if (IsEqualGUID(sample_type->subtype, WMMEDIASUBTYPE_Y160)) {
			res = S_FALSE; // No conversion necessary.
		}
		else {
			// Unable to convert this format to Y16.
			res = E_FAIL;
		}
	}


	_ASSERT(res == S_OK || res == S_FALSE);
	return res;
}

#if 0
static HRESULT ConvertSample_X8R8G8B8_Y8(const EVFormat* dfmt, void* dst, void* src) {
	HRESULT res = S_OK;
	
	res = E_NOTIMPL;
	
	_ASSERT(res == S_OK);
	return res;
}

static HRESULT ConvertSample_X8R8G8B8_Y16(const EVFormat* dfmt, void* dst, void* src) {
	HRESULT res = S_OK;
	
	res = E_NOTIMPL;
	
	_ASSERT(res == S_OK);
	return res;
}

__forceinline unsigned __int8 clip(int x) {
	if (x < 0) return 0;
	if (x > 255) return 255;
	return x;
}

static HRESULT ConvertSample_X8R8G8B8_I420(const EVFormat* dfmt, void* dst, void* src) {
	HRESULT res = S_OK;
	
	long rows = dfmt->height;
	long cols = dfmt->width;
	long half_rows = (rows + 1) / 2;
	long half_cols = (cols + 1) / 2;

	unsigned __int32* dst_ui = reinterpret_cast<unsigned __int32*>(dst);

	const unsigned __int8* src_y = reinterpret_cast<const unsigned __int8*>(src);
	const unsigned __int8* src_cb = src_y + rows * cols;
	const unsigned __int8* src_cr = src_cb + half_rows * half_cols;

	for (long row = 0; row < rows; ++row) {
		unsigned __int32* dst_row = dst_ui + row * cols;
		const unsigned __int8* src_y_row = src_y + row * cols;
		const unsigned __int8* src_cb_row = src_cb + row / 2 * half_cols;
		const unsigned __int8* src_cr_row = src_cr + row / 2 * half_cols;

		for (long col = 0; col < dfmt->width; ++col) {

			const int y = src_y_row[col] - 16;
			const int cb = src_cb_row[col/2] - 128;
			const int cr = src_cr_row[col/2] - 128;

			unsigned __int8 r = clip((298 * y            + 409 * cr + 128) >> 8);
			unsigned __int8 g = clip((298 * y - 100 * cb - 208 * cr + 128) >> 8);
			unsigned __int8 b = clip((298 * y + 516 * cb            + 128) >> 8);
			
			dst_row[col] = (0xFF << 24) | (r << 16) | (g << 8) | (b << 0);
		}
	}

	_ASSERT(res == S_OK);
	return res;
}

static HRESULT ConvertSample_Y8_X8R8G8B8(const EVFormat* dfmt, void* dst, void* src) {
	HRESULT res = S_OK;
	
	res = E_NOTIMPL;
	
	_ASSERT(res == S_OK);
	return res;
}

static HRESULT ConvertSample_Y8_Y16(const EVFormat* dfmt, void* dst, void* src) {
	HRESULT res = S_OK;
	
	unsigned __int8* dst_us = reinterpret_cast<unsigned __int8*>(dst);
	const unsigned __int16* src_ub = reinterpret_cast<const unsigned __int16*>(src);

	for (long r = 0; r < dfmt->height; ++r) {
		unsigned __int8* dst_row = dst_us + r * dfmt->width;
		const unsigned __int16* src_row = src_ub + r * dfmt->width;
		for (long c = 0; c < dfmt->width; ++c) {
			dst_row[c] = src_row[c] >> 8;
		}
	}

	_ASSERT(res == S_OK);
	return res;
}

static HRESULT ConvertSample_Y16_X8R8G8B8(const EVFormat* dfmt, void* dst, void* src) {
	HRESULT res = S_OK;
	
	res = E_NOTIMPL;
	
	_ASSERT(res == S_OK);
	return res;
}

static HRESULT ConvertSample_Y16_Y8(const EVFormat* dfmt, void* dst, void* src) {
	HRESULT res = S_OK;
	
	unsigned __int16* dst_us = reinterpret_cast<unsigned __int16*>(dst);
	const unsigned __int8* src_ub = reinterpret_cast<const unsigned __int8*>(src);

	for (long r = 0; r < dfmt->height; ++r) {
		unsigned __int16* dst_row = dst_us + r * dfmt->width;
		const unsigned __int8* src_row = src_ub + r * dfmt->width;
		for (long c = 0; c < dfmt->width; ++c) {
			dst_row[c] = src_row[c] << 8;
		}
	}

	_ASSERT(res == S_OK);
	return res;
}
#endif