#include "common.hpp"
#include "DSInputSink.hpp"
#include "Utility/ArrayEnumMediaTypes.hpp"
EQUIS_USING_NS1(equis);

DSInputSink::DSInputSink(DSInput* parent, HRESULT& res)
: BaseSink(res), m_parent(parent),
	m_frame_event(NULL), m_frame(NULL),
	m_unused(NULL),
	m_callback(NULL), m_callback_user(NULL)
{
	if (res == S_OK) {
		m_frame_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		if (!m_frame_event) {
			res = HRESULT_FROM_WIN32(GetLastError());
		}
	}
}

DSInputSink::~DSInputSink() {
}

HRESULT DSInputSink::SetDesired(const EVFormat* format, const AM_MEDIA_TYPE* type) {
	m_desired_format = *format;
	m_desired_type = type;
	return S_OK;
}

HRESULT DSInputSink::SetCallback(EVICallbackProc callback, void* user) {
	HRESULT res = S_OK;

	MutexLock lock(&m_lock);
	if (res == S_OK) {
		res = lock.Enter(INFINITE);
	}

	if (res == S_OK) {
		m_callback = callback;
		m_callback_user = user;
	}

	return res;
}

HRESULT DSInputSink::PushUnusedFrame(DSInputFrame* frame) {
	HRESULT res = S_OK;
	if (frame == NULL) res = E_POINTER;

	if (res == S_OK) {
		// Ensure the sample information is cleared so that DirectShow can reuse it.
		frame->SetSample(NULL, NULL);
	}

	MutexLock lock(&m_lock);
	if (res == S_OK) {
		res = lock.Enter(INFINITE);
	}

	if (res == S_OK) {
		frame->next = m_unused;
		m_unused = frame;
	}

	return res;
}

HRESULT DSInputSink::ReadFrame(DWORD timeout, DSInputFrame** frame) {
	HRESULT res = S_OK;

	if (frame == NULL) res = E_POINTER;
	else *frame = NULL;

	MutexLock lock(&m_lock);
	if (res == S_OK) {
		HANDLE handles[] = { m_lock.GetHandle(), m_frame_event };
		DWORD waitResult = WaitForMultipleObjects(
			EQUIS_ARRAY_CAPACITY(handles), handles, TRUE, timeout);
		if (waitResult == WAIT_OBJECT_0) lock.SetEntered();
		else if (waitResult == WAIT_TIMEOUT) res = E_ABORT;
		else res = E_FAIL;
	}

	if (res == S_OK) {
		// Grab the current frame. The event is automatically reset.
		*frame = m_frame;
	}

	return res;
}


STDMETHODIMP DSInputSink::ChangingConnectionType(const AM_MEDIA_TYPE* type) {
	HRESULT res = QueryAccept(type);
	if (res == S_FALSE) res = VFW_E_TYPE_NOT_ACCEPTED;

	if (res == S_OK) {
		m_actual_type = type;
	}

	return res;
}

#define MINTTF (10000000LL/50)
STDMETHODIMP DSInputSink::QueryAccept(const AM_MEDIA_TYPE *type) {
	if (type == NULL) return E_POINTER;
	return MatchesMediaType(m_desired_type.ptr(), type) ? S_OK : S_FALSE;
}

STDMETHODIMP DSInputSink::EnumMediaTypes(IEnumMediaTypes **types) {
	if (types == NULL) return E_POINTER;

	*types = new ArrayEnumMediaTypes(m_desired_type.ptr(), 1);
	return S_OK;
}

STDMETHODIMP DSInputSink::Receive(IMediaSample *sample) {
	HRESULT res = S_OK;

	if (sample == NULL) res = E_POINTER;

	MutexLock lock(&m_lock);
	if (res == S_OK) {
		res = lock.Enter(100);
		// Do not block excessively here, to avoid stalling the DirectShow
		// thread during shutdown. If acquiring the lock times out, simply
		// return with success. Only the current frame will be dropped.
		if (res == E_ABORT) {
			return S_OK;
		}
	}

	DSInputFrame* frame = NULL;
	if (res == S_OK) {
		frame = m_unused;
		if (frame) {
			m_unused = frame->next;
		}
		else {
			// No frame object is available for this sample, create one.
			frame = new DSInputFrame(m_parent);
		}
		frame->SetFormat(&m_desired_format);
		frame->SetSample(m_actual_type.ptr(), sample);
	}

	m_frame = frame;
	if (!SetEvent(m_frame_event)) {
		res = HRESULT_FROM_WIN32(GetLastError());
	}
	if (m_callback) {
		m_callback(m_frame, m_callback_user);
	}

	_ASSERT(res == S_OK);
	return res;
}
