#include "common.hpp"
#include "ArrayEnumPins.hpp"
#include "BaseSink.hpp"
EQUIS_USING_NS1(equis);

BaseSink::BaseSink(HRESULT &res)
: m_lock(res), m_startTime(0), m_state(State_Stopped), m_flushing(false)
{
    memset(m_name, 0, sizeof(m_name));
    m_pins[0] = this;
}

BaseSink::~BaseSink()
{
}

STDMETHODIMP BaseSink::QueryInterface(REFIID iid, void** result) {
    HRESULT res = BaseUnknown::QueryInterface(iid, result);
    if (res == E_NOINTERFACE) {
        if (IsEqualIID(iid, IID_IPersist)) {
            *result = implicit_cast<IPersist *>(this);
        } else if (IsEqualIID(iid, IID_IMediaFilter)) {
            *result = implicit_cast<IMediaFilter *>(this);
        } else if (IsEqualIID(iid, IID_IBaseFilter)) {
            *result = implicit_cast<IBaseFilter *>(this);
        } else if (IsEqualIID(iid, IID_IPin)) {
            *result = implicit_cast<IPin *>(this);
        } else if (IsEqualIID(iid, IID_IMemInputPin)) {
            *result = implicit_cast<IMemInputPin *>(this);
        }
        if (*result) {
            AddRef();
            res = S_OK;
        }
    }
    return res;
}

STDMETHODIMP BaseSink::GetClassID(CLSID *clsid)
{
    return E_FAIL;
}

STDMETHODIMP BaseSink::Stop()
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);
    res = lock.Enter(INFINITE);
    if (SUCCEEDED(res))
    {
        m_state = State_Stopped;
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::Pause()
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);
    res = lock.Enter(INFINITE);
    if (SUCCEEDED(res)) {
        m_state = State_Paused;
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::Run(REFERENCE_TIME startTime)
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);
    res = lock.Enter(INFINITE);
    if (SUCCEEDED(res)) {
        m_state = State_Running;
        m_startTime = startTime;
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::GetState(DWORD timeout, FILTER_STATE *result)
{
    HRESULT res = S_OK;

    if (result == NULL) {
        res = E_POINTER;
    } else {
        MutexLock lock(&m_lock);
        res = lock.Enter(timeout);
        if (SUCCEEDED(res)) {
            *result = m_state;
        }
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::SetSyncSource(IReferenceClock *clock)
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);
    res = lock.Enter(INFINITE);
    if (SUCCEEDED(res)) {
        m_clock = clock;
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::GetSyncSource(IReferenceClock **clock)
{
    HRESULT res = S_OK;

    if (clock == NULL) {
        res = E_POINTER;
    } else {
        MutexLock lock(&m_lock);
        res = lock.Enter(INFINITE);
        if (SUCCEEDED(res)) {
            *clock = m_clock.Get();
        }
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::EnumPins(IEnumPins **enumPins)
{
    HRESULT res = S_OK;

    if (enumPins == NULL) {
        res = E_POINTER;
    } else {
        *enumPins = new ArrayEnumPins(m_pins, 1);
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::FindPin(LPCWSTR id, IPin **pin)
{
    _ASSERT(false); return E_NOTIMPL;
}

STDMETHODIMP BaseSink::JoinFilterGraph(IFilterGraph *graph, LPCWSTR name)
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);
    res = lock.Enter(INFINITE);
    if (SUCCEEDED(res)) {
        if (m_state != State_Stopped) {
            res = VFW_E_NOT_STOPPED;
        } else {
            m_graph.AttachWithoutAddRefOrRelease(graph);

            memset(m_name, 0, sizeof(m_name));
            if (name != NULL) {
                wcscpy_s(m_name, name);
            }
        }
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::QueryFilterInfo(FILTER_INFO *info)
{
    HRESULT res = S_OK;

    if (info == NULL) {
        res = E_POINTER;
    } else {
        MutexLock lock(&m_lock);
        res = lock.Enter(INFINITE);
        if (SUCCEEDED(res)) {
            memcpy(info->achName, m_name, sizeof(m_name));
            info->pGraph = m_graph.Get();
        }
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::QueryVendorInfo(LPWSTR *vendorInfo)
{
    _ASSERT(false); return E_NOTIMPL;
}

STDMETHODIMP BaseSink::Connect(IPin *targetPin, const AM_MEDIA_TYPE *mediaType)
{
    _ASSERT(false); return E_NOTIMPL;
}

STDMETHODIMP BaseSink::ReceiveConnection(IPin *sourcePin, const AM_MEDIA_TYPE *type)
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);
    res = lock.Enter(INFINITE);
    if (SUCCEEDED(res)) {
        if (m_state != State_Stopped) {
            res = VFW_E_NOT_STOPPED;
        } else if (m_sourcePin.IsNotNull()) {
            res = VFW_E_ALREADY_CONNECTED;
        } else {
            res = ChangingConnectionType(type);
            if (res == S_OK) {
                m_connectionMediaType = type;
                m_sourcePin.Attach(sourcePin);
            }
        }
    }
    _ASSERT(SUCCEEDED(res) || res == VFW_E_TYPE_NOT_ACCEPTED);
    return res;
}

STDMETHODIMP BaseSink::Disconnect()
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);
    res = lock.Enter(INFINITE);
    if (SUCCEEDED(res)) {
        if (m_state != State_Stopped) {
            res = VFW_E_NOT_STOPPED;
        } else if (m_sourcePin.IsNull()) {
            res = S_FALSE;
        } else {
            m_connectionMediaType.clear();
            m_sourcePin.Detach();
        }
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::ConnectedTo(IPin **pin)
{
    HRESULT res = S_OK;

    if (pin == NULL) {
        res = E_POINTER;
    } else {
        MutexLock lock(&m_lock);
        res = lock.Enter(INFINITE);
        if (SUCCEEDED(res)) {
            *pin = m_sourcePin.Get();
        }
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::ConnectionMediaType(AM_MEDIA_TYPE *type)
{
    HRESULT res = S_OK;

    if (type == NULL) {
        res = E_POINTER;
    } else {
        MutexLock lock(&m_lock);
        res = lock.Enter(INFINITE);
        if (SUCCEEDED(res)) {
            res = CopyMediaType(type, m_connectionMediaType.ptr());
        }
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::QueryPinInfo(PIN_INFO *pinInfo)
{
    HRESULT res = S_OK;

    if (pinInfo == NULL) {
        res = E_POINTER;
    } else {
        AddRef();
        pinInfo->pFilter = implicit_cast<IBaseFilter*>(this);
        pinInfo->dir = PINDIR_INPUT;
        wcscpy_s(pinInfo->achName, L"Input");
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::QueryDirection(PIN_DIRECTION *pinDirection)
{
    HRESULT res = S_OK;

    if (pinDirection == NULL) {
        res = E_POINTER;
    } else {
        *pinDirection = PINDIR_INPUT;
    }
    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP BaseSink::QueryId(LPWSTR *id)
{
    HRESULT res = S_OK;

    if (id == NULL) {
        res = E_POINTER;
    } else {
        static const WCHAR my_id[] = L"0";

        LPWSTR new_id = static_cast<LPWSTR>(
            CoTaskMemAlloc(sizeof(my_id)));
        if (!new_id) {
            res = E_OUTOFMEMORY;
        } else {
            memcpy(new_id, my_id, sizeof(my_id));
        }
    }

    return res;
}

STDMETHODIMP BaseSink::QueryInternalConnections(IPin **pins, ULONG *length)
{
    return E_NOTIMPL;
}

STDMETHODIMP BaseSink::EndOfStream()
{
    _ASSERT(false); return E_NOTIMPL;
}

STDMETHODIMP BaseSink::BeginFlush()
{
    m_flushing = true;
    return S_OK;
}

STDMETHODIMP BaseSink::EndFlush()
{
    HRESULT res = S_OK;

    MutexLock lock(&m_lock);

    if (res == S_OK) {
        res = lock.Enter(INFINITE);
    }
    if (res == S_OK) {
        m_flushing = false;
    }
    return S_OK;
}

STDMETHODIMP BaseSink::NewSegment(REFERENCE_TIME startTime, REFERENCE_TIME stopTime, double rate)
{
    _ASSERT(false); return E_NOTIMPL;
}

STDMETHODIMP BaseSink::GetAllocator(IMemAllocator **allocator)
{
    return VFW_E_NO_ALLOCATOR;
}

STDMETHODIMP BaseSink::NotifyAllocator(IMemAllocator *allocator, BOOL readOnly)
{
    HRESULT res = S_OK;

    ALLOCATOR_PROPERTIES request, actual;
    res = allocator->GetProperties(&actual);
    if (res == S_OK) {
        if (actual.cBuffers < 2 || actual.cbAlign < 16) {
			request.cBuffers = std::max(4L, actual.cBuffers);
			request.cbBuffer = actual.cbBuffer;
			request.cbAlign = std::max(16L, actual.cbAlign);
			request.cbPrefix = actual.cbPrefix;

			res = allocator->SetProperties(&request, &actual);
			if (res == S_OK) {
				if (actual.cBuffers < 2 || actual.cbAlign < 16) {
					// Failed to acquire acceptable properties.
					res = E_FAIL;
				}
			}
        }
    }

    _ASSERT(res == S_OK);
    return res;
}

STDMETHODIMP BaseSink::GetAllocatorRequirements(ALLOCATOR_PROPERTIES *properties)
{
    if (properties == NULL) {
        return E_POINTER;
    } else {
        properties->cBuffers = 4;
        properties->cbBuffer = m_connectionMediaType->lSampleSize;
        properties->cbAlign = 16;
        properties->cbPrefix = 0;
        return S_OK;
    }
}

STDMETHODIMP BaseSink::ReceiveMultiple(IMediaSample** samples, LONG capacity, LONG* length)
{
    HRESULT res = S_OK;

    if (samples == NULL || (capacity > 1 && length == NULL)) {
        res = E_POINTER;
    }

    // Multiple samples may take some time to process. This function does not hold
    // the lock, so that other operations may be interleaved with the repeated
    // Receives.

    if (SUCCEEDED(res)) {
        LONG index = 0;
        for (; index < capacity; ++index) {
            res = Receive(samples[index]);
            if (res != S_OK) break;
        }
        if (length) *length = index;
    }

    _ASSERT(SUCCEEDED(res));
    return res;
}

STDMETHODIMP
BaseSink::ReceiveCanBlock() {
    return S_OK;
}
