#include "common.hpp"
#include "EnumVideoCaptures.hpp"
EQUIS_USING_NS1(equis);

EnumVideoCaptures::EnumVideoCaptures(HRESULT &res) {
    VariantInit(&m_current_name);
    VariantInit(&m_current_path);

    ComPtr<ICreateDevEnum> factory;
    if (res == S_OK) {
        res = CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC_SERVER,
            factory.GetUuid(), factory.DetachAndGetVoidPointer());

		if (res == CO_E_NOTINITIALIZED) {
			// Transparently initialize COM.
			res = CoInitialize(NULL);
			if (res == S_OK) {
				// Retry the object creation once.
				res = CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC_SERVER,
					factory.GetUuid(), factory.DetachAndGetVoidPointer());
			}
		}
    }

    if (res == S_OK) {
        res = factory->CreateClassEnumerator(
            CLSID_VideoInputDeviceCategory,
            m_source.DetachAndGetPointer(),
            0);
    }

	_ASSERT(res == S_OK);
}

EnumVideoCaptures::~EnumVideoCaptures() {
    VariantClear(&m_current_name);
    VariantClear(&m_current_path);
}

STDMETHODIMP EnumVideoCaptures::QueryInterface(REFIID iid, void** result) {
    return BaseUnknown::QueryInterface(iid, result);
}

STDMETHODIMP EnumVideoCaptures::Next() {
    HRESULT res = S_OK;

    m_current.Detach();
    m_properties.Detach();

    if (m_source.IsNull())
    {
        // No device of the specified type was found.
        res = S_FALSE;
    } else {
        ULONG fetched;
        res = m_source->Next(1, m_current.DetachAndGetPointer(), &fetched);

        if (res == S_OK) {
            res = m_current->BindToStorage(NULL, NULL,
                IID_IPropertyBag, m_properties.DetachAndGetVoidPointer());
        }
    }

    _ASSERT(res == S_OK || res == S_FALSE);
    return res;
}

STDMETHODIMP EnumVideoCaptures::GetName(LPCWSTR* result)
{
    HRESULT res = S_OK;

    if (result == NULL) {
        res = E_POINTER;
    } else if (m_properties.IsNull()) {
        res = E_FAIL;
    } else {
        VariantClear(&m_current_name);

        res = m_properties->Read(L"FriendlyName", &m_current_name, NULL);
        if (res == S_OK) {
            if (m_current_name.vt != VT_BSTR) {
                // Internal error: property bag produced an unexpected type.
                // TODO: should use VariantChangeType to do the conversion.
                res = E_FAIL;
            } else {
                *result = m_current_name.bstrVal;
            }
        }
    }

    _ASSERT(res == S_OK);
    return res;
}

STDMETHODIMP EnumVideoCaptures::GetPath(LPCWSTR* result)
{
    HRESULT res = S_OK;

    if (result == NULL) {
        res = E_POINTER;
    } else if (m_properties.IsNull()) {
        res = E_FAIL;
    } else {
		*result = NULL;
        VariantClear(&m_current_name);

        res = m_properties->Read(L"DevicePath", &m_current_name, NULL);
		if (res == S_OK) {
            if (m_current_name.vt != VT_BSTR) {
                // Internal error: property bag produced an unexpected type.
                // TODO: should use VariantChangeType to do the conversion.
                res = E_FAIL;
            } else {
                *result = m_current_name.bstrVal;
            }
        }
		else if (res == E_INVALIDARG || res == HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND)) {
			// Point Gray appears to register a capture device,
			// but behaves incorrectly when the device path is queried.
			// This should be E_INVALIDAR
			res = E_INVALIDARG;
		}
    }

    _ASSERT(res == S_OK || res == E_INVALIDARG);
    return res;
}

STDMETHODIMP EnumVideoCaptures::BindToObject(REFIID iid, void** object)
{
    HRESULT res = S_OK;

    if (object == NULL) {
        res = E_POINTER;
    } else {
        res = m_current->BindToObject(NULL, NULL, iid, object);
    }

    _ASSERT(res == S_OK);
    return res;
}

HRESULT EnumVideoCaptures::BindPathToObject(LPCWSTR path, REFIID iid, void** object)
{
    HRESULT res = S_OK;

    EnumVideoCaptures captures(res);
    while (res == S_OK) {
        res = captures.Next();

        LPCWSTR current_path = NULL;
        if (res == S_OK) {
            res = captures.GetPath(&current_path);
        }

        if (res == S_OK && (path == NULL || wcscmp(path, current_path) == 0)) {
            res = captures.BindToObject(iid, object);
            break;
        }
    }

    _ASSERT(res == S_OK || res == S_FALSE);
    return res;
}
