#include "common.hpp"
#include "format.h"

static bool compatible(EVFormat* desired, const EVFormat* candidate, int fuzz);
static bool better(const EVFormat* candidate, const EVFormat* current);
static unsigned short best_type(const EVFormat* desired, const EVFormat* provided);

long evfChoose(
	EVFormat* desired,
	const EVFormat* formats, long end)
{
	// If no formats are specified, no compatible one is present.
	if (!formats || end <= 0) return -1;

	// If the 'desired' format is NULL, use a local object with reasonable defaults.
	// This object will be modified by the later code, so must not be static or const.
	EVFormat default_desired = { 0, 0, 0, 0, 0 };
	if (desired == NULL) desired = &default_desired;

	if (desired->types == 0) {
		// Assume RGB32 if no types are actually requested.
		desired->types = EVF_TYPE_A8R8G8B8;
	}

	// Remove abstract and undefined flags from the types.
	desired->types &= (EVF_TYPE_A8R8G8B8 | EVF_TYPE_Y8 | EVF_TYPE_Y16);

	const EVFormat* best = NULL;

	for (int fuzz = 0; fuzz < 3; ++fuzz) {
		for (long i = 0; i < end; ++i) {
			const EVFormat* candidate = &formats[i];

			if (compatible(desired, candidate, fuzz) && better(candidate, best)) {
				best = candidate;
			}
		}
	}

	if (best == NULL) return -1;

	desired->types = best_type(desired, best);
	desired->flags = best->flags;
	desired->width = best->width;
	desired->height = best->height;
	desired->usecs = best->usecs;
	return best - formats;
}

static bool compatible(EVFormat* desired, const EVFormat* candidate, int fuzz) {

	unsigned short best_area = desired->flags & EVF_FLAG_BEST_AREA;
	unsigned short best_rate = desired->flags & EVF_FLAG_BEST_RATE;

	bool types_ok;
	if (fuzz <= 0) { // Accept only overlapping types.
		types_ok = (desired->types & candidate->types) != 0;
	}
	else if (fuzz == 1) { // Accept types with compatible colorspaces
		types_ok |=
			(desired->types & EVF_TYPE_COLOR_MASK) != 0 && 
			(candidate->types & EVF_TYPE_COLOR_MASK) != 0;
		types_ok |=
			(desired->types & EVF_TYPE_GRAY_MASK) != 0 && 
			(candidate->types & EVF_TYPE_GRAY_MASK) != 0;
	}
	else { // Accept any source and destination types
		types_ok |= desired->types != 0;
		types_ok |= candidate->types != 0;
	}

	bool width_ok = (desired->width == 0);
	if (best_area) width_ok |= desired->width <= candidate->width;
	else width_ok |= desired->width == candidate->width;

	bool height_ok = (desired->height == 0);
	if (best_area) height_ok |= desired->height <= candidate->height;
	else height_ok |= desired->height == candidate->height;

	bool rate_ok = (desired->usecs == 0);
	if (best_rate) rate_ok |= desired->usecs >= candidate->usecs;
	else rate_ok |= desired->usecs == candidate->usecs;

	return types_ok && width_ok && height_ok && rate_ok;
}

static bool better(const EVFormat* candidate, const EVFormat* current) {
	if (current == NULL) return true;

	if (current->usecs > candidate->usecs) return true;
	if (current->usecs < candidate->usecs) return false;

	long candidate_area = candidate->width * candidate->height;
	long current_area = current->width * current->height;

	if (candidate_area > current_area) return true;
	if (candidate_area < current_area) return false;

	if (candidate->width > current->width) return true;
	if (candidate->width < current->width) return false;

	// No difference in resolution or rate, so keep the earlier format.
	return false;
}

static unsigned short best_type(const EVFormat* desired, const EVFormat* provided) {
	
	unsigned short shared_types = desired->types & provided->types;
	if (shared_types & EVF_TYPE_A8R8G8B8) return EVF_TYPE_A8R8G8B8;
	else if (shared_types & EVF_TYPE_Y16) return EVF_TYPE_Y16;
	else if (shared_types & EVF_TYPE_Y8) return EVF_TYPE_Y8;

	if ((desired->types & EVF_TYPE_COLOR_MASK) != 0 &&
		(provided->types & EVF_TYPE_COLOR_MASK) != 0)
	{
		// Want and have color formats.
		if (desired->types & EVF_TYPE_A8R8G8B8) return EVF_TYPE_A8R8G8B8;
	}

	if ((desired->types & EVF_TYPE_GRAY_MASK) != 0 &&
		(provided->types & EVF_TYPE_GRAY_MASK) != 0)
	{
		// Want and have grayscale formats.
		if (desired->types & EVF_TYPE_Y8) return EVF_TYPE_Y8;
		if (desired->types & EVF_TYPE_Y16) return EVF_TYPE_Y16;
	}

	// No convenient matches, try for any destination type.
	if (desired->types & EVF_TYPE_A8R8G8B8) return EVF_TYPE_A8R8G8B8;
	else if (desired->types & EVF_TYPE_Y16) return EVF_TYPE_Y16;
	else if (desired->types & EVF_TYPE_Y8) return EVF_TYPE_Y8;

	// Unknown or invalid type specified.
	return 0;
}