#include "common.hpp"
#include "format.h"
#include "common/image/image.hpp"
EQUIS_USING_NS1(equis);

static long convert_direct(
	EVFormat* dfmt, void* dst, long dcap,
	const EVFormat* sfmt, const void* src, long send);

static long convert_modify(
	EVFormat* dfmt, void* dst, long dcap,
	const EVFormat* sfmt, const void* src, long send);

long evfConvert(
	EVFormat* dfmt, void* dst, long dcap,
	const EVFormat* sfmt, const void* src, long send)
{
	if (!dfmt || !sfmt) return -1;

	// Do not currently support resizing images.
	if ((dfmt->width != 0 && dfmt->width != sfmt->width) ||
		(dfmt->height != 0 && dfmt->height != sfmt->height)) {
		return -1;
	}

	// Do not currently support in-place conversions.
	if (src && dst == src) {
		return -1;
	}

	// The source must have a concrete type to perform the conversion.
	if (dst && count_one_bits(sfmt->types) != 1) {
		return -1;
	}

	if (dfmt->types & sfmt->types) {
		// Found a common type. Do not need to modify the pixel data.
		return convert_direct(dfmt, dst, dcap, sfmt, src, send);
	}

	return convert_modify(dfmt, dst, dcap, sfmt, src, send);
}


static long convert_direct(
	EVFormat* dfmt, void* dst, long dcap,
	const EVFormat* sfmt, const void* src, long send)
{
	unsigned short shared_types = dfmt->types & sfmt->types;
	unsigned short vflip = (dfmt->flags ^ sfmt->flags) & EVF_FLAG_BOTTOM_UP;

	// Choose the most capable shared type.
	if (shared_types & EVF_TYPE_A8R8G8B8) dfmt->types = EVF_TYPE_A8R8G8B8;
	else if (shared_types & EVF_TYPE_Y16) dfmt->types = EVF_TYPE_Y16;
	else if (shared_types & EVF_TYPE_Y8) dfmt->types = EVF_TYPE_Y8;
	else return -1;

	dfmt->width = sfmt->width;
	dfmt->height = sfmt->height;
	dfmt->usecs = sfmt->usecs;

	long stride = evfStride(dfmt);
	if (stride < 0) return -1;

	long size = stride * dfmt->height;
	if (dst) {
		// Missing buffer or insufficient image sizes for format.
		if (!src || dcap < size || send < size) return -1;

		if (!vflip) {
			memcpy(dst, src, size);
		} else {
			for (long i = 0, j = dfmt->height - 1; j >= 0; ++i, --j) {
				memcpy(
					byte_offset(dst, stride * j),
					byte_offset(src, stride * i),
					stride);
			}
		}
	}

	return size;
}

static long convert_modify(
	EVFormat* dfmt, void* dst, long dcap,
	const EVFormat* sfmt, const void* src, long send)
{
	bool vflip = ((dfmt->flags ^ sfmt->flags) & EVF_FLAG_BOTTOM_UP) != 0;

	// For this function to be invoked, there must be no common types
	// between the source and destination formats.

	if ((dfmt->types & EVF_TYPE_A8R8G8B8) != 0 && (sfmt->types & EVF_TYPE_COLOR_MASK) != 0) {
		// Use RGB32 when the source includes color information.
		dfmt->types = EVF_TYPE_A8R8G8B8;
	}
	else if ((dfmt->types & EVF_TYPE_Y8) != 0 && (sfmt->types & EVF_TYPE_GRAY_MASK) != 0) {
		// Use Y8 if no color information is present.
		// If the caller wants 16-bpp grayscale, they need to ask for it.
		dfmt->types = EVF_TYPE_Y8;
	}
	// No good matches found. Pick a single destination format.
	else if (dfmt->types & EVF_TYPE_A8R8G8B8) dfmt->types = EVF_TYPE_A8R8G8B8;
	else if (dfmt->types & EVF_TYPE_Y8) dfmt->types = EVF_TYPE_Y8;
	else if (dfmt->types & EVF_TYPE_Y16) dfmt->types = EVF_TYPE_Y16;
	else return -1;

	long width = dfmt->width = sfmt->width;
	long height = dfmt->height = sfmt->height;
	dfmt->usecs = sfmt->usecs;

	long dst_stride = evfStride(dfmt);
	if (dst_stride < 0) return -1;

	long dst_size = dst_stride * dfmt->height;
	if (dst) {
		// Missing buffer or insufficient image sizes for format.
		if (!src || dcap < dst_size || send < evfSize(sfmt)) return -1;

		// Identity conversions are handled by convert_direct.

#define COMB(D,S) (((D)<<16)|(S))
#define CASE(D,S) \
	case COMB(EVF_TYPE_ ## D, EVF_TYPE_ ## S): image::convertImage<image::D,image::S>( \
		dst, src, width, height, vflip); break;

		switch (COMB(dfmt->types, sfmt->types)) {
			CASE(Y8, Y16);
			CASE(Y8, A8R8G8B8);
			CASE(Y16, Y8);
			CASE(Y16, A8R8G8B8);
			CASE(A8R8G8B8, Y8);
			CASE(A8R8G8B8, Y16);
			default: return -1;
		}
#undef COMB
#undef CASE
	}

	return dst_size;
}
