#include "common.hpp"
#include "xna-common.hpp"
#include "Input_Ftir.hpp"

EQUIS_USING_NS3(Equis,Xna,Input);

FtirEvent::FtirEvent(int type, float x, float y, float mass)
: m_type(type), m_position(x, y), m_mass(mass)
{
}

FtirEvent::FtirEvent(int type, Vector2 position, float mass)
: m_type(type), m_position(position), m_mass(mass)
{
}

Vector2 FtirEvent::Position::get()
{
	return m_position;
}

float FtirEvent::X::get()
{
	return m_position.X;
}

float FtirEvent::Y::get()
{
	return m_position.Y;
}

bool FtirEvent::Press::get()
{
	return m_type > 0;
}

bool FtirEvent::Release::get()
{
	return m_type < 0;
}

float FtirEvent::Mass::get()
{
	return m_mass;
}

int FtirEvent::SessionId::get()
{
	return m_session_id;
}

int FtirEvent::CursorId::get()
{
	return m_cursor_id;
}

int FtirEvent::Frame::get()
{
	return m_frame;
}

int FtirEvent::Since::get()
{
	return m_since;
}

FtirEvent ^FtirEvent::FromNative(ETEvent *native)
{
	FtirEvent^ e = gcnew FtirEvent(native->type, native->x, native->y, native->mass);
	e->m_frame = native->frame;
	e->m_since = native->since;
	e->m_session_id = native->session_id;
	e->m_cursor_id = native->cursor_id;
	return e;
}

void FtirEvent::ToNative(ETEvent *native)
{
	if (native == nullptr) {
		Marshal::ThrowExceptionForHR(E_POINTER);
	}
	native->x = X;
	native->x = Y;
	native->mass = Mass;
}

#if 0
Ftir ^Ftir::FromCapture(VideoDevice ^captureDevice)
{
	return FromCapture(captureDevice->Path);
}
#endif

Ftir ^Ftir::FromCapture(String ^capturePath)
{
	if (capturePath == nullptr) {
		ETSurface* surface = etCreate();
		if (etSourceVideoOpen(surface, NULL) < 0) {
			etDelete(surface);
			Marshal::ThrowExceptionForHR(E_FAIL);
		}
		return gcnew Ftir(surface);
	}

	Marshal::ThrowExceptionForHR(E_NOTIMPL);

	IntPtr capturePathAnsi;
	try {
		capturePathAnsi = Marshal::StringToCoTaskMemAnsi(capturePath);

		ETSurface* surface = etCreate();
		
		if (etSourceVideoOpen(surface, reinterpret_cast<LPSTR>(capturePathAnsi.ToPointer())) < 0) {
			etDelete(surface);
			Marshal::ThrowExceptionForHR(E_FAIL);
		}

		return gcnew Ftir(surface);
	}
	finally {
		if (capturePathAnsi != IntPtr::Zero) Marshal::FreeCoTaskMem(capturePathAnsi);
	}
}

Ftir ^Ftir::FromFile(String ^filePath)
{
	Marshal::ThrowExceptionForHR(E_NOTIMPL);
	return nullptr;
}

VideoFormat ^Ftir::Format::get()
{
	HRESULT res;

	EVFormat nativeFormat;

	res = etGetFormat(m_source, &nativeFormat);
	Marshal::ThrowExceptionForHR(res);

	return VideoFormat::FromNative(&nativeFormat);
}

float Ftir::RealFramesPerSecond::get()
{
	float fps = 0.0f;
	//Marshal::ThrowExceptionForHR(
		//m_source->GetRealFramesPerSecond(&fps));
	return fps;
}

#if 0
Video ^Ftir::InputPreview::get()
{
	Marshal::ThrowExceptionForHR(E_NOTIMPL);
	return nullptr;
}

Video ^Ftir::ChangePreview::get()
{
	Marshal::ThrowExceptionForHR(E_NOTIMPL);
	return nullptr;
}
#endif

static HRESULT GetComPtr(Object ^managed, REFIID iid, void **result)
{
	FieldInfo ^field = managed->GetType()->GetField("pComPtr", BindingFlags::Instance | BindingFlags::NonPublic);
	IUnknown *unknown = (IUnknown *) Pointer::Unbox(field->GetValue(managed));
	return unknown->QueryInterface(iid, result);
}

void Ftir::PreviewInput(Texture2D^ texture) {
	ComPtr<IDirect3DTexture9> texture_d3d;
	Marshal::ThrowExceptionForHR(GetComPtr(texture,
		texture_d3d.GetUuid(), texture_d3d.DetachAndGetVoidPointer()));

	D3DLOCKED_RECT rect;
	Marshal::ThrowExceptionForHR(texture_d3d->LockRect(
		0, &rect, NULL, D3DLOCK_DISCARD));

	try {
		etPreview(m_source, ET_PREVIEW_INPUT, rect.pBits, rect.Pitch, rect.Pitch * 480 * 4);
	}
	finally {
		texture_d3d->UnlockRect(0);
	}
}

void Ftir::PreviewFiltered(Texture2D^ texture) {
	ComPtr<IDirect3DTexture9> texture_d3d;
	Marshal::ThrowExceptionForHR(GetComPtr(texture,
		texture_d3d.GetUuid(), texture_d3d.DetachAndGetVoidPointer()));

	D3DLOCKED_RECT rect;
	Marshal::ThrowExceptionForHR(texture_d3d->LockRect(
		0, &rect, NULL, D3DLOCK_DISCARD));

	try {
		etPreview(m_source, ET_PREVIEW_FILTERED, rect.pBits, rect.Pitch, rect.Pitch * 480 * 4);
	}
	finally {
		texture_d3d->UnlockRect(0);
	}
}

void Ftir::Start()
{
}

void Ftir::Stop()
{
}

void Ftir::Calibrate(array<Vector2> ^desired, array<Vector2> ^reported)
{
	if (desired == nullptr || reported == nullptr) {
		etCalibrate(m_source, NULL, NULL, 0);
		return;
	}

	equis::array<float> desired_elems;
	equis::array<float> reported_elems;

	long count = min(desired->Length, reported->Length);
	if (!desired_elems.resize(2*count) || !reported_elems.resize(2*count)) {
		Marshal::ThrowExceptionForHR(E_OUTOFMEMORY);
	}

	for (long i = 0; i < count; ++i) {
		desired_elems[2*i+0] = desired[i].X;
		desired_elems[2*i+1] = desired[i].Y;
		reported_elems[2*i+0] = reported[i].X;
		reported_elems[2*i+1] = reported[i].Y;
	}

	etCalibrate(m_source, desired_elems.begin(), reported_elems.begin(), 2 * count);
}

array<FtirEvent ^> ^Ftir::PollEvents(float requiredMass)
{
	const LONG nativesLength = 64;
	ETEvent natives[nativesLength];

	long natives_end = etPoll(m_source, natives, EQUIS_ARRAY_CAPACITY(natives));
	if (natives_end < 0) return nullptr;

	for (; natives_end > 0; --natives_end)
	{
		if (Math::Abs(natives[natives_end - 1].mass) >= requiredMass) break;
	}

	array<FtirEvent ^> ^events = gcnew array<FtirEvent ^>(natives_end);
	for (LONG i = 0; i < natives_end; ++i)
	{
		events[i] = FtirEvent::FromNative(&natives[i]);
	}
	return events;
}

Ftir::Ftir(ETSurface *source)
{
	m_source = source;
}

Ftir::~Ftir(void)
{
}
