#include "common.hpp"
#include "xna-common.hpp"
#include "Input_Video.hpp"

EQUIS_USING_NS3(Equis,Xna,Input);

VideoFormat::VideoFormat(SurfaceFormat format, int width, int height, LONGLONG frameInterval)
: m_format(format), m_width(width), m_height(height), m_frameInterval(frameInterval)
{
}

SurfaceFormat VideoFormat::Format::get()
{
	return m_format;
}

int VideoFormat::Width::get()
{
	return m_width;
}

int VideoFormat::Height::get()
{
	return m_height;
}

LONGLONG VideoFormat::FrameInterval::get()
{
	return m_frameInterval;
}

float VideoFormat::FramesPerSecond::get()
{
	return 1.0e7f / float(m_frameInterval);
}

struct FormatD3DToXna
{
	D3DFORMAT d3d;
	SurfaceFormat xna;

	FormatD3DToXna(D3DFORMAT d3d, SurfaceFormat xna)
		: d3d(d3d), xna(xna)
	{}

	FormatD3DToXna(LPCSTR fourcc, SurfaceFormat xna)
		: d3d((D3DFORMAT)MAKEFOURCC(fourcc[0], fourcc[1], fourcc[2], fourcc[3])), xna(xna)
	{}
};

#define F(d3d, xna) FormatD3DToXna((d3d), (xna))

//
// WARNING: The XNA SurfaceFormat enumeration is big-endian, so is in the
//          reverse order from the D3DFORMAT!
//
// Most of these are probably correct, but mileage may vary. Formats like
// UYVY and YUY2 are particularly likely to be broken.
// 
static FormatD3DToXna formatD3DToXna[] = {

	// *** All SurfaceFormat equivalences

    F(D3DFMT_A8, SurfaceFormat::Alpha8),
	F(D3DFMT_R3G3B2, SurfaceFormat::Bgr233),
    F(D3DFMT_R8G8B8, SurfaceFormat::Bgr24),
    F(D3DFMT_X8R8G8B8, SurfaceFormat::Bgr32),
    F(D3DFMT_X4R4G4B4, SurfaceFormat::Bgr444),
    F(D3DFMT_X1R5G5B5, SurfaceFormat::Bgr555),
    F(D3DFMT_R5G6B5, SurfaceFormat::Bgr565),
    F(D3DFMT_A2R10G10B10, SurfaceFormat::Bgra1010102),
    F(D3DFMT_A8R3G3B2, SurfaceFormat::Bgra2338),
    F(D3DFMT_A4R4G4B4, SurfaceFormat::Bgra4444),
    F(D3DFMT_A1R5G5B5, SurfaceFormat::Bgra5551),
    F(D3DFMT_A8R8G8B8, SurfaceFormat::Color),
    F(D3DFMT_D15S1, SurfaceFormat::Depth15Stencil1),
    F(D3DFMT_D16, SurfaceFormat::Depth16),
    F(D3DFMT_D24X8, SurfaceFormat::Depth24),
    F(D3DFMT_D24X4S4, SurfaceFormat::Depth24Stencil4),
    F(D3DFMT_D24S8, SurfaceFormat::Depth24Stencil8),
    F(D3DFMT_D24FS8, SurfaceFormat::Depth24Stencil8Single),
    F(D3DFMT_D32, SurfaceFormat::Depth32),
    F(D3DFMT_DXT1, SurfaceFormat::Dxt1),
    F(D3DFMT_DXT2, SurfaceFormat::Dxt2),
    F(D3DFMT_DXT3, SurfaceFormat::Dxt3),
    F(D3DFMT_DXT4, SurfaceFormat::Dxt4),
    F(D3DFMT_DXT5, SurfaceFormat::Dxt5),
    F(D3DFMT_R16F, SurfaceFormat::HalfSingle),
    F(D3DFMT_G16R16F, SurfaceFormat::HalfVector2),
    F(D3DFMT_A16B16G16R16F, SurfaceFormat::HalfVector4),
    F(D3DFMT_L16, SurfaceFormat::Luminance16),
    F(D3DFMT_L8, SurfaceFormat::Luminance8),
    F(D3DFMT_A8L8, SurfaceFormat::LuminanceAlpha16),
    F(D3DFMT_A4L4, SurfaceFormat::Luminance8),
    F(D3DFMT_MULTI2_ARGB8, SurfaceFormat::Multi2Bgra32),
	F(D3DFMT_A2W10V10U10, SurfaceFormat::NormalizedAlpha1010102),
	F(D3DFMT_V8U8, SurfaceFormat::NormalizedByte2),
	F(D3DFMT_CxV8U8, SurfaceFormat::NormalizedByte2Computed),
	F(D3DFMT_Q8W8V8U8, SurfaceFormat::NormalizedByte4),
	F(D3DFMT_L6V5U5, SurfaceFormat::NormalizedLuminance16),
	F(D3DFMT_X8L8V8U8, SurfaceFormat::NormalizedLuminance32),
	F(D3DFMT_V16U16, SurfaceFormat::NormalizedShort2),
	F(D3DFMT_Q16W16V16U16, SurfaceFormat::NormalizedShort4),
    F(D3DFMT_P8, SurfaceFormat::Palette8),
    F(D3DFMT_A8P8, SurfaceFormat::PaletteAlpha16),
    F(D3DFMT_G16R16, SurfaceFormat::Rg32),
    F(D3DFMT_X8B8G8R8, SurfaceFormat::Rgb32),
    F(D3DFMT_A2B10G10R10, SurfaceFormat::Rgba1010102),
    F(D3DFMT_A8B8G8R8, SurfaceFormat::Rgba32),
    F(D3DFMT_A16B16G16R16, SurfaceFormat::Rgba64),
    F(D3DFMT_R32F, SurfaceFormat::Single),
    F(D3DFMT_G32R32F, SurfaceFormat::Vector2),
    F(D3DFMT_A32B32G32R32F, SurfaceFormat::Vector4),
    F(D3DFMT_R8G8_B8G8, SurfaceFormat::VideoGrGb),
    F(D3DFMT_G8R8_G8B8, SurfaceFormat::VideoRgBg),
	F(D3DFMT_UYVY, SurfaceFormat::VideoYuYv),
	F(D3DFMT_YUY2, SurfaceFormat::VideoUyVy),

	// All SurfaceFormat values must be defined above. As the first match
	// is used, SurfaceFormats may be reused below for extended D3DFORMAT
	// values.

	// *** FOURCC codes that do not have macros.
	//
	// These do not have real SurfaceFormat equivalents, so should specify
	// the format to which they can most efficiently be converted. For
	// example: I420 starts with an 8-bit monochrome plane, so can easily
	// be converted to Luminance8.
	//
	// Add entries here to avoid an assertion during the conversion. If no
	// reasonable equivalent is available, use SurfaceFormat::Unknown.

	F("GREY", SurfaceFormat::Luminance8),
	F("I420", SurfaceFormat::Luminance8),
	F("Y8  ", SurfaceFormat::Luminance8),
	F("Y800", SurfaceFormat::Luminance8),

	// Both values Unknown marks the end of the array.
	F(D3DFMT_UNKNOWN, SurfaceFormat::Unknown)
};

#undef F

VideoFormat ^VideoFormat::FromNative(EVFormat *native)
{
	if (native == NULL) {
		Marshal::ThrowExceptionForHR(E_POINTER);
	}

#if 0
	SurfaceFormat surfaceFormat = SurfaceFormat::Unknown;
	for (FormatD3DToXna *convert = formatD3DToXna;
		 convert->d3d != D3DFMT_UNKNOWN || convert->xna != SurfaceFormat::Unknown;
		 ++convert)
	{
		if (native->types == convert->d3d) {
			surfaceFormat = convert->xna;
			break;
		}
	}
	_ASSERT(surfaceFormat != SurfaceFormat::Unknown);
#endif


	return gcnew VideoFormat(
		SurfaceFormat::Color, // TODO: Detect color from native format
		native->width,
		native->height,
		native->usecs);
}

void VideoFormat::ToNative(EVFormat *native)
{
	if (native == NULL) {
		Marshal::ThrowExceptionForHR(E_POINTER);
	}

	D3DFORMAT d3dFormat = D3DFMT_UNKNOWN;
	for (FormatD3DToXna *convert = formatD3DToXna;
		 convert->d3d != D3DFMT_UNKNOWN || convert->xna != SurfaceFormat::Unknown;
		 ++convert)
	{
		if (Format == convert->xna) {
			d3dFormat = convert->d3d;
			break;
		}
	}
	_ASSERT(d3dFormat != D3DFMT_UNKNOWN);

	native->types = d3dFormat;
	native->width = m_width;
	native->height = m_height;
	native->usecs = static_cast<long>(m_frameInterval / 10);
}

String ^VideoFormat::ToString()
{
	return
		Format.ToString() + " " +
		Width + "x" +
		Height + " " +
		(1e7 / FrameInterval).ToString("F0") + " fps";
}
